#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

# Import checking code - Check if we can actually
# import in the pygtk stuff and report an intelligeable
# error to the user if we can't
import sys, os

from intl import _

failure = None

try:
	import gtk, GtkExtra
except ImportError, strerror:
	failure = strerror

# Import the appropriate GTK constant file. In the win32 pygtk
# package, GTK is GTKconst
if os.name == "nt":
	try:
		import GTKconst
	except ImportError, strerror:
		failure = strerror
else:
	try:
		import GTK
	except ImportError, strerror:
		failure = strerror

if failure is not None:
	sys.stderr.write (_("ftpcube - Compilation Error: %s.\n") %failure)
	sys.stderr.write (_("\tUnable to import the pygtk modules.\n"))
	sys.stderr.write (_("\tFtpCube cannot run without pygtk modules properly installed.\n"))
	sys.stderr.write (_("\tFor more information on how to obtain these vital packages, visit:\n"))
	sys.stderr.write (_("\t\thttp://ftpcube.sourceforge.net\n"))
	sys.exit (1)

__all__ = [
	'aboutwin',
	'app',
	'bookmarkwin',
	'cmdwin',
	'connectwin',
	'ctrlwin',
	'defaults',
	'filewin',
	'ftp',
	'logging',
	'main',
	'optionwin',
	'threads',
	'intl',
	'cache',
	'browsewin'
]



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk
import sys, os
import main

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class OptionWindow:

	def __init__ (self):

		self.opts = main.app.load_config (main.app.get_config_file ())

		self.dialog = gtk.GtkDialog ()
		self.dialog.set_title (_("ftpcube - Options"))
		self.dialog.set_policy (gtk.FALSE, gtk.FALSE, gtk.TRUE)
		self.dialog.set_position (gtk.WIN_POS_CENTER)
		icon_pic, icon_mask = gtk.create_pixmap_from_xpm (self.dialog, None, os.path.join (main.ICONS_PREFIX, "ftpcube.xpm"))
		self.dialog.set_icon (gtk.GtkPixmap (icon_pic, icon_mask))

		# Make ourselves modal
		gtk.grab_add (self.dialog)

		packer = gtk.GtkPacker ()
		packer.show ()
		self.dialog.action_area.add (packer)
		table = gtk.GtkTable (2, 1)
		table.set_col_spacing (0, 100)
		table.show ()
		packer.add (table)

		button = gtk.GtkButton (_("Accept"))
		button.set_usize (50, 25)
		button.connect ("clicked", self.accept)
		table.attach (button, 0, 1, 0, 1, xoptions=GTK.EXPAND)
		button.show ()

		button = gtk.GtkButton (_("Cancel"))
		button.set_usize (50, 25)
		button.connect ("clicked", self.cancel)
		table.attach (button, 1, 2, 0, 1, xoptions=GTK.EXPAND)
		button.show ()

		self.notebook = gtk.GtkNotebook ()
		self.notebook.set_tab_pos (gtk.POS_TOP)
		self.dialog.vbox.add (self.notebook)
		self.notebook.show ()

		# Create the tabs and add them and show them
		self.connection_tab = ConnectionTab (self.opts)
		label = gtk.GtkLabel (_("Connections"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.connection_tab.get_page (), label)

		self.logging_tab = LoggingTab (self.opts)
		label = gtk.GtkLabel (_("Logging"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.logging_tab.get_page (), label)

		self.colors_tab = ColorsTab (self.opts)
		label = gtk.GtkLabel (_("Colors"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.colors_tab.get_page (), label)

#		self.external_tab = ExternalTab (self.opts)
#		label = gtk.GtkLabel (_("Externals"))
#		label.set_padding (2, 2)
#		label.show ()
#		self.notebook.append_page (self.external_tab.get_page (), label)
#
		self.dialog.show ()

	def accept (self, button):

		# Need to do this to make sure the current color is saved
		self.colors_tab.save_current_color ()

		options = self.get_options ()
		self.opts.update (options)
		main.app.save_config (main.app.get_config_file (), self.opts)
		main.app.update (self.opts)

		# Set the logging parameter for each active thread in case it
		# changed
		for thread in main.app.get_thread_list ():
			thread.set_log_active (self.opts['logging'])

		self.cancel (button)

	def cancel (self, button):

		self.dialog.hide ()
		self.dialog.destroy ()

	def get_options (self):

		opts = { }
		opts.update (self.connection_tab.get_options ())
		opts.update (self.logging_tab.get_options ())
		opts.update (self.colors_tab.get_options ())
		#opts.update (self.external_tab.get_options ())
		return opts

class ConnectionTab:

	def __init__ (self, opts):

		self.page = gtk.GtkFrame (_("Default Connection Settings"))
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		vbox = gtk.GtkVBox ()
		vbox.show ()
		self.page.add (vbox)

		frame = gtk.GtkFrame (_("Log Files"))
		frame.set_border_width (10)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.show ()
		vbox.pack_start (frame, expand=GTK.FALSE)
		packer = gtk.GtkPacker ()
		packer.show ()
		frame.add (packer)
		table = gtk.GtkTable (3, 4)
		table.set_row_spacings (7)
		table.show ()
		packer.add (table, options=GTK.FILL_X+GTK.FILL_Y, pad_x=15, pad_y=15)

		# Create the timeout slot
		label = gtk.GtkLabel (_("Connection timeout:"))
		label.show ()
		table.attach (label, 0, 1, 0, 1)
		self.timeout = gtk.GtkEntry ()
		if opts is not None and opts['timeout']:
			self.timeout.set_text ("%s" %opts['timeout'])
		self.timeout.show ()
		table.attach (self.timeout, 1, 2, 0, 1)
		label = gtk.GtkLabel (_("(seconds)"))
		label.show ()
		table.attach (label, 2, 3, 0, 1)

		label = gtk.GtkLabel (_("Idle Timeout:"))
		label.show ()
		table.attach (label, 0, 1, 1, 2)
		self.main_idle = gtk.GtkEntry ()
		if opts is not None and opts['main_idle']:
			self.main_idle.set_text ("%s" %opts['main_idle'])
		self.main_idle.show ()
		table.attach (self.main_idle, 1, 2, 1, 2)
		label = gtk.GtkLabel (_("(seconds)"))
		label.show ()
		table.attach (label, 2, 3, 1, 2)

		# Create the delay slot
		label = gtk.GtkLabel (_("Delay between retries:"))
		label.show ()
		table.attach (label, 0, 1, 2, 3)
		self.delay = gtk.GtkEntry ()
		if opts is not None and opts['delay']:
			self.delay.set_text ("%s" %opts['delay'])
		self.delay.show ()
		table.attach (self.delay, 1, 2, 2, 3)
		label = gtk.GtkLabel (_("(seconds)"))
		label.show ()
		table.attach (label, 2, 3, 2, 3)

		# Create the retries slot
		label = gtk.GtkLabel (_("Number of attempts:"))
		label.show ()
		table.attach (label, 0, 1, 3, 4)
		self.retries = gtk.GtkEntry ()
		if opts is not None and opts['retries']:
			self.retries.set_text ("%s" %opts['retries'])
		self.retries.show ()
		table.attach (self.retries, 1, 2, 3, 4)

		# Create the thread options frame
		frame = gtk.GtkFrame (_("Thread options"))
		frame.set_border_width (10)
		frame.set_shadow_type (GTK.SHADOW_ETCHED_OUT)
		frame.show ()
		vbox.pack_start (frame, expand=GTK.FALSE)
		packer = gtk.GtkPacker ()
		packer.show ()
		frame.add (packer)
		table = gtk.GtkTable (3, 1)
		table.show ()
		packer.add (table, options=GTK.FILL_X+GTK.FILL_Y, pad_x=15, pad_y=15)

		# Create the idle slot
		label = gtk.GtkLabel (_("Idle timeout for threads:"))
		label.show ()
		table.attach (label, 0, 1, 0, 1)
		self.thread_idle = gtk.GtkEntry ()
		self.thread_idle.set_usize (40, -1)
		if opts is not None and opts['thread_idle']:
			self.thread_idle.set_text ("%s" %opts['thread_idle'])
		self.thread_idle.show ()
		table.attach (self.thread_idle, 1, 2, 0, 1)
		label = gtk.GtkLabel (_("(seconds)"))
		label.show ()
		table.attach (label, 2, 3, 0, 1)

	def get_options (self):

		opts = { }
		if self.timeout.get_text ():
			opts['timeout'] = int (self.timeout.get_text ())
		if self.main_idle.get_text ():
			opts['main_idle'] = int (self.main_idle.get_text ())
		if self.delay.get_text ():
			opts['delay'] = int (self.delay.get_text ())
		if self.retries.get_text ():
			opts['retries'] = int (self.retries.get_text ())
		if self.thread_idle.get_text ():
			opts['thread_idle'] = int (self.thread_idle.get_text ())
		return opts

	def get_page (self):
		return self.page

class LoggingTab:

	def __init__ (self, opts):

		# Create instance variables
		self.log = 0

		self.page = gtk.GtkFrame (_("FTP Logging"))
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		vbox = gtk.GtkVBox ()
		vbox.show ()
		self.page.add (vbox)

		frame = gtk.GtkFrame (_("Activate Logging"))
		frame.set_border_width (10)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.show ()
		vbox.pack_start (frame, expand=GTK.FALSE)
		table = gtk.GtkTable (2, 1)
		table.show ()
		frame.add (table)

		self.activate_button = gtk.GtkRadioButton (None, _("Turn ON Logging"))
		self.activate_button.connect ("clicked", self.activate_clicked)
		self.activate_button.show ()
		self.deactivate_button = gtk.GtkRadioButton (self.activate_button, _("Turn OFF Logging"))
		self.deactivate_button.connect ("clicked", self.deactivate_clicked)
		self.deactivate_button.show ()
		if opts is not None:
			if int (opts['logging']):
				self.activate_button.set_active (gtk.TRUE)
			else:
				self.deactivate_button.set_active (gtk.TRUE)
		packer = gtk.GtkPacker ()
		packer.show ()
		packer.add (self.activate_button)
		table.attach (packer, 0, 1, 0, 1, ypadding=4)
		packer = gtk.GtkPacker ()
		packer.show ()
		packer.add (self.deactivate_button)
		table.attach (packer, 1, 2, 0, 1, ypadding=4)

		packer = gtk.GtkPacker ()
		packer.show ()
		vbox.pack_start (packer)
		frame = gtk.GtkFrame (_("Log Files"))
		frame.set_border_width (10)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.show ()
		packer.add (frame, options=GTK.FILL_X)
		packer = gtk.GtkPacker ()
		packer.show ()
		frame.add (packer)
		table = gtk.GtkTable (2, 3)
		table.set_row_spacings (7)
		table.show ()
		packer.add (table, options=GTK.FILL_X+GTK.FILL_Y, pad_x=25, pad_y=25)

		label = gtk.GtkLabel (_("FTP command log:"))
		label.show ()
		self.cmds = gtk.GtkEntry ()
		if opts is not None and opts['cmd_log']:
			self.cmds.set_text (opts['cmd_log'])
		self.cmds.show ()
		table.attach (label, 0, 1, 0, 1)
		table.attach (self.cmds, 1, 2, 0, 1)

		label = gtk.GtkLabel (_("Download log:"))
		label.show ()
		self.downloads = gtk.GtkEntry ()
		if opts is not None and opts['download_log']:
			self.downloads.set_text (opts['download_log'])
		self.downloads.show ()
		table.attach (label, 0, 1, 1, 2)
		table.attach (self.downloads, 1, 2, 1, 2)

		label = gtk.GtkLabel (_("Upload log:"))
		label.show ()
		self.uploads = gtk.GtkEntry ()
		if opts is not None and opts['upload_log']:
			self.uploads.set_text (opts['upload_log'])
		self.uploads.show ()
		table.attach (label, 0, 1, 2, 3)
		table.attach (self.uploads, 1, 2, 2, 3)

	def activate_clicked (self, button):
		self.log = 1

	def deactivate_clicked (self, button):
		self.log = 0

	def get_page (self):
		return self.page

	def get_options (self):

		return {
			'logging' : self.log,
			'cmd_log' : self.cmds.get_text (),
			'download_log' : self.downloads.get_text (),
			'upload_log'   : self.uploads.get_text (),
		}

class ColorsTab:

	def __init__ (self, opts):

		self.page = gtk.GtkFrame (_("Colors"))
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		vbox = gtk.GtkVBox ()
		vbox.show ()
		self.page.add (vbox)
		self.color_select = gtk.GtkColorSelection ()
		self.color_select.show ()
		vbox.pack_start (self.color_select)

		packer = gtk.GtkPacker ()
		packer.show ()
		vbox.pack_start (packer)
		frame = gtk.GtkFrame (_("Text Type"))
		frame.show ()
		packer.add (frame, pad_x=5, pad_y=5)
		hbox = gtk.GtkHBox ()
		hbox.show ()
		frame.add (hbox)

		self.local_button = gtk.GtkRadioButton (None, _("Local Text"))
		self.local_button.connect ("clicked", self.local_clicked)
		self.local_button.show ()
		self.remote_button = gtk.GtkRadioButton (self.local_button, _("Remote Text"))
		self.remote_button.connect ("clicked", self.remote_clicked)
		self.remote_button.show ()
		self.error_button = gtk.GtkRadioButton (self.local_button, _("Error Text"))
		self.error_button.connect ("clicked", self.error_clicked)
		self.error_button.show ()
		hbox.pack_start (self.local_button, padding=10)
		hbox.pack_start (self.remote_button, padding=10)
		hbox.pack_start (self.error_button, padding=10)

		# Set current values and load the local as default
		if opts.has_key ('local_color'):
			self.local_color = opts['local_color']
		if opts.has_key ('remote_color'):
			self.remote_color = opts['remote_color']
		if opts.has_key ('error_color'):
			self.error_color = opts['error_color']
		self.set_current_color (self.local_color)

	def set_current_color (self, color):

		# Assign the color twice so we set the current color,
		# as well as the memory slot
		self.color_select.set_color (color)
		self.color_select.set_color (color)

	def save_current_color (self):

		if self.local_button.get_active ():
			self.local_color = self.color_select.get_color ()
		elif self.remote_button.get_active ():
			self.remote_color = self.color_select.get_color ()
		elif self.error_button.get_active ():
			self.error_color = self.color_select.get_color ()

	def local_clicked (self, button):

		if self.local_button.get_active ():
			self.set_current_color (self.local_color)
		else:
			self.local_color = self.color_select.get_color ()

	def remote_clicked (self, button):

		if self.remote_button.get_active ():
			self.set_current_color (self.remote_color)
		else:
			self.remote_color = self.color_select.get_color ()

	def error_clicked (self, button):

		if self.error_button.get_active ():
			self.set_current_color (self.error_color)
		else:
			self.error_color = self.color_select.get_color ()

	def get_page (self):
		return self.page

	def get_options (self):

		return {
			'local_color' : self.local_color,
			'remote_color' : self.remote_color,
			'error_color' : self.error_color,
		}

class ExternalTab:

	def __init__ (self, opts):

		self.page = gtk.GtkFrame (_("External Helpers"))
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		vbox = gtk.GtkVBox ()
		vbox.show ()
		self.page.add (vbox)
		packer = gtk.GtkPacker ()
		packer.show ()
		vbox.pack_start (packer, expand=GTK.FALSE, fill=GTK.FALSE)
		table = gtk.GtkTable (2, 2)
		table.set_row_spacings (7)
		table.show ()
		packer.add (table, options=GTK.FILL_X, pad_x=15, pad_y=15)

		label = gtk.GtkLabel (_("Launching Apps:"))
		label.show ()
		table.attach (label, 0, 1, 0, 1)
		self.launcher = gtk.GtkEntry ()
		if opts is not None and opts['launcher']:
			self.launcher.set_text (opts['launcher'])
		self.launcher.show ()
		table.attach (self.launcher, 1, 2, 0, 1)

		label = gtk.GtkLabel (_("Viewer:"))
		label.show ()
		table.attach (label, 0, 1, 1, 2)
		self.viewer = gtk.GtkEntry ()
		if opts is not None and opts['viewer']:
			self.viewer.set_text (opts['viewer'])
		self.viewer.show ()
		table.attach (self.viewer, 1, 2, 1, 2)

	def get_options (self):

		return {
			'launcher' : self.launcher.get_text (),
			'viewer'   : self.viewer.get_text (),
		}

	def get_page (self):
		return self.page



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import string # For backwards compatibility with 1.5.x
import sys, os
import main, app, ftp, threads

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class CtrlWin:

	"""The control window object

	The control window object provides the framework for displaying and creating
	the ftp control objects, such as queuing and threading methods. These objects
	handle user display tasks and high-level manipulation of the ftp backend.
	The control window displays all of the status information concerning actual
	transfers and queuing of transfers. The CtrlWin object provides a consilated
	interface to the lower level sub-objects, each of which are responsible for
	displaying a control "tab" and implementing its related functions.

	There should only be one instance of the control window and it is provided
	a reference to its container from the Application object upon creation."""

	def __init__ (self, container):
		"""Creates the CtrlWin object and displays its contents within the
		container box provided by the Application object. The sub objects
		that occupy the tabs of the notebook are also created."""

		box = gtk.GtkVBox ()
		box.set_usize (275, -1)
		box.show ()
		container.add (box)

		self.notebook = gtk.GtkNotebook ()
		self.notebook.set_tab_pos (gtk.POS_TOP)
		box.pack_start (self.notebook)
		self.notebook.show ()

		status_box = gtk.GtkHBox ()
		status_box.set_usize (-1, 15)
		status_box.show ()
		box.pack_start (status_box, expand=gtk.FALSE)
		self.status_init (status_box)

		self.queue_tab = QueueTab (self)
		label = gtk.GtkLabel (_("Queue"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.queue_tab.get_page (), label)

		self.threads_tab = ThreadsTab (self)
		label = gtk.GtkLabel (_("Threads"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.threads_tab.get_page (), label)

		self.download_tab = DownloadTab (self)
		label = gtk.GtkLabel (_("Downloads"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.download_tab.get_page (), label)

		self.failure_tab = FailureTab (self)
		label = gtk.GtkLabel (_("Failures"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.failure_tab.get_page (), label)

		# Default to the thread page
		self.notebook.set_page (1)

	def status_init (self, container):
		"""Initializes the status bar within the status 'container', which
		displays information about the queuing process. The label describes
		how many items remain in the queue, and the progress bar indicates
		how many queuing items have been processed as a percentage of the
		batch queueing."""

		table = gtk.GtkTable (2, 1)
		table.show ()
		container.pack_start (table, padding=10)

		packer = gtk.GtkPacker ()
		packer.show ()
		table.attach (packer, 0, 1, 0, 1)
		self.queue_label = gtk.GtkLabel (_("0 Queued"))
		self.queue_label.show ()
		packer.add (self.queue_label, anchor=GTK.ANCHOR_WEST)

		self.queue_progress = gtk.GtkProgressBar ()
		self.queue_progress.set_usize (-1, 7)
		table.attach (self.queue_progress, 1, 2, 0, 1, yoptions=GTK.EXPAND)

		# We default to having no progress out of no items
		self.progress = (0, 0)

	def status_queued_add (self, queued):
		"""Updates the queuing status bar to reflect that one more item
		has been added to the queue and displays the total number of items
		'queued' remaining in the queue."""

		cur, max = self.progress
		max = max + 1
		self.progress = (cur, max)
		self.queue_progress.show ()
		self.queue_progress.update (float (cur) / float (max))
		self.queue_label.set_text (_("%d Queued") %queued)

	def status_queued_remove (self, queued):
		"""Updates the queueing status bar to reflect that an item has
		just been removed and is being processed and displays the total
		number of items 'queued' remaining in the queue."""

		if queued == 0:
			self.status_queue_clear ()
		else:
			cur, max = self.progress
			cur = cur + 1
			self.progress = (cur, max)
			self.queue_progress.update (float (cur) / float (max))
			self.queue_label.set_text (_("%d Queued") %queued)

	def status_queue_clear (self):
		"""Updates to the status bar to reflect that the queue is empty."""

		self.queue_progress.hide ()
		self.progress = (0, 0)
		self.queue_label.set_text (_("0 Queued"))

	def add_to_queue (self, hash):
		"""Provides a high level interface to add a list of files from an
		FTP site into the download queue. Each file entry in the
		hash['files'] key is a tuple of (name, written size, date, flags,
		true size)."""

		for name, size, date, flags, truesize in hash['files']:
			self.queue_tab.add_to_queue ({
				'file'        : name,
				'server'      : hash['server'],
				'port'        : hash['port'],
				'username'    : hash['username'],
				'password'    : hash['password'],
				'remote_path' : hash['remote_path'],
				'local_path'  : hash['local_path'],
				'size'        : truesize,
				'attempt'     : 1,
				'maxattempts' : hash['maxattempts'],
				'direction'   : hash['direction'],
				'flags'       : flags,
				'method'      : hash['method'],
			})

	def add_to_downloads (self, file):
		"""Adds a file to the completed downloads tab."""

		self.download_tab.add_to_list (file)

	def add_to_failure (self, hash):
		"""Adds the structure stored in the dictionary into the failures
		tab."""

		self.failure_tab.add_to_list (hash)

	def attach_main_thread (self, thread):
		"""Places the main thread object 'thread' into the list of threads
		in the ThreadTab object for display."""

		self.threads_tab.attach_main_thread (thread)

	def get_thread_list (self):
		"""Returns a list of active threads from the ThreadTab."""

		return self.threads_tab.get_thread_list ()

class QueueTab:

	"""The Queueing Tab

	Displays a list of the files waiting in the queue and provides several
	buttons for manipulating the processing of the queue, such as a pausing
	button, a clearing button, and a save/load button."""

	headings = [ _('Host'), _('Filename'), _('Size'), _('Attempt'), _('Direction'), _('Port'),
	             _('Username'), _('Password'), _('Remote Path'), _('Local Path'), _('Type'),
	             _('Method') ]

	def __init__ (self, ctrl):
		"""Creates the queueing tab and sets up the appropriate widgets. Also
		initializaes the queue to empty. 'ctrl' is a reference to the parent
		CtrlWin object."""

		# Initialize instance variables
		self.ctrlwin = ctrl
		self.disabled = 0                # For enabling and disabling the queueing process

		self.page = gtk.GtkVBox ()
		self.page.show ()

		box = gtk.GtkHBox ()
		box.show ()
		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.set_space_size (10)
		toolbar.set_space_style (GTK.TOOLBAR_SPACE_LINE)
		toolbar.show ()
		box.pack_start (toolbar, padding=6)
		self.page.pack_start (box, expand=gtk.FALSE, padding=4)

		enable_pic, enable_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "thread_idle.xpm"))
		clear_pic, clear_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "clear.xpm"))
		load_pic, load_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "load_queue.xpm"))
		save_pic, save_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "save_queue.xpm"))

		toolbar.append_element (GTK.TOOLBAR_CHILD_TOGGLEBUTTON, None,
			"", _("Enable Queuing Process"), _("Enable Queuing Process"),
			gtk.GtkPixmap (enable_pic, enable_mask), self.enable_queuing_process)
		toolbar.append_space ()
		toolbar.append_item ("", _("Clear Queue"), _("Clear Queue"),
			gtk.GtkPixmap (clear_pic, clear_mask),
			self.clear_queue)
		toolbar.append_space ()
		toolbar.append_item ("", _("Load Queue"), _("Load Queue"),
			gtk.GtkPixmap (load_pic, load_mask),
			self.load_queue)
		toolbar.append_item ("", _("Save Queue"), _("Save Queue"),
			gtk.GtkPixmap (save_pic, save_mask),
			self.save_queue)

		win = gtk.GtkScrolledWindow ()
		win.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		win.show ()
		self.page.pack_start (win)

		# Create the list
		self.list = gtk.GtkCList (len (self.headings), self.headings)
		self.list.connect ("select_row", self.select_items)
		self.list.set_column_width (0, 80)
		self.list.set_column_width (1, 150)
		self.list.set_column_width (2, 80)
		self.list.set_column_width (3, 60)
		# Make the invisible ones truly invisible
		for i in range (4,12):
			self.list.set_column_visibility (i, 0)
		self.list.set_selection_mode (GTK.SELECTION_MULTIPLE)
		self.list.show ()
		win.add_with_viewport (self.list)

	def enable_queuing_process (self, button):
		"""Toggles pausing/enabling the queuing process. All entries simply wait in
		the queue while the process is disabled."""

		if not self.disabled:
			self.disabled = 1
		else:
			self.disabled = 0

	def add_to_queue (self, hash):
		"""Adds a single file with attributes in dictionary 'hash' into the queue
		and updates the status bar accordingly."""

		self.list.append ((hash['server'],
		                   hash['file'],
		                   str (hash['size']),
		                   "%s/%s" %(hash['attempt'], hash['maxattempts']),
		                   str (hash['direction']),
		                   str (hash['port']),
		                   hash['username'],
		                   hash['password'],
		                   hash['remote_path'],
		                   hash['local_path'],
		                   hash['flags'],
		                   str (hash['method'])))
		self.ctrlwin.status_queued_add (self.list.rows)
		if not self.disabled:
			self.ctrlwin.threads_tab.check_if_available_threads ()

	def get_next (self):
		"""Returns a tuple containing the data in the next entry of the list."""

		if not self.length ():
			return None

		# Build the first row entry
		entry = [ ]
		for i in range (self.list.columns):
			entry.append (self.list.get_text (0, i))
		return tuple (entry)

	def remove_next (self):
		"""Removes the next queued entry from the list, signaling that the entry
		is being processed."""

		self.list.freeze ()
		self.list.remove (0)
		self.list.thaw ()
		self.ctrlwin.status_queued_remove (self.list.rows)

	def length (self):
		"""Returns the number of entries in the list."""

		# Check if we're disabled and return a zero length to calling
		# functions will think we're empty
		if self.disabled:
			return 0
		else:
			return self.list.rows

	def clear_queue (self, button):
		"""Purge all entries from the queue."""

		self.list.clear ()
		self.ctrlwin.status_queue_clear ()

	def save_queue (self, button):
		"""Saves the contents of the queue to a file, given in the
		pop-up file selection box."""

		path = GtkExtra.file_sel_box (title=_("ftpcube - Save queue file"))
		if path is None:
			return

		try:
			file = open (path, 'w')
		except IOError, strerror:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ] ))
			return

		for r in range (self.list.rows):
			entry = [ ]
			for c in range (self.list.columns):
				entry.append (self.list.get_text (r, c))
			entry = string.join (entry, '|')
			file.write ("%s\n" %entry)
		file.close ()

	def load_queue (self, button):
		"""Loads queue entries from a queuing file, provided by the
		user in a pop-up file selection box."""

		path = GtkExtra.file_sel_box (_("ftpcube - Load queue file"))
		if path is None:
			return

		try:
			file = open (path, "r")
		except IOError, strerror:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ] ))
			return

		line = file.readline ()
		while line:
			entry = string.split (line, '|')
			self.list.append (list (entry))
			line = file.readline ()
		file.close ()

	def select_items (self, list, r, c, event):
		pass

	def get_page (self):
		return self.page

class ThreadsTab:

	"""The threading tab object

	The threading tab object manages the actions of the download threads and displays
	their status. A spin button is displayed to allow the user to select the maximum
	number of threads to use for transferring and another indicates whether
	multi-threaded transferring should be used at all.

	The thread tab implements an idle function which performs updates on the
	status of each thread at specified intervals. The idle function determines
	if a thread is free and feeds it an item from the queue, and also updates idle
	counts."""

	def __init__ (self, ctrl):
		"""Creates a thread tab object and uses 'ctrl' as a container for the thread
		status display. Also sets up the idle timer to execute the idle function every
		second to update thread status and check if further processing is needed."""

		# Initialize instance variables
		self.threads = [ ]
		self.main_thread = None
		self.main_transfer = 0
		self.ctrlwin = ctrl

		self.page = gtk.GtkVBox ()
		self.page.show ()

		box = gtk.GtkHBox ()
		box.show ()
		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.set_space_size (10)
		toolbar.set_space_style (GTK.TOOLBAR_SPACE_LINE)
		toolbar.show ()
		box.pack_start (toolbar, padding=6)
		self.page.pack_start (box, expand=gtk.FALSE, padding=4)

		box = gtk.GtkHBox ()
		toolbar.append_widget (box, "", "")
		box.show ()
		self.no_threads = gtk.GtkSpinButton (gtk.GtkAdjustment (0, 0, 16, 1, 10, 0), 1, 0)
		self.no_threads.set_numeric (gtk.TRUE)
		self.no_threads.show ()
		box.pack_start (self.no_threads)
		label = gtk.GtkLabel (_("max threads"))
		label.show ()
		box.pack_start (label)

		toolbar.append_space ()
		pic, mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "main_thread.xpm"))
		toolbar.append_element (GTK.TOOLBAR_CHILD_TOGGLEBUTTON, None,
			"", _("Only use main thread for transfers"),
			_("Only use main thread for transfers"),
			gtk.GtkPixmap (pic, mask),
			self.use_main_thread)

		self.transfer_label = gtk.GtkLabel ()
		self.update_transfer_totals (0, 0.0)
		self.transfer_label.set_line_wrap (gtk.TRUE)
		self.transfer_label.show ()
		self.page.pack_start (self.transfer_label, expand=gtk.FALSE)

		# Create the page as a scrollable window so multiple threads don't
		# spill outside the screen
		vbox = gtk.GtkVBox ()
		vbox.show ()
		hbox = gtk.GtkHBox ()
		hbox.show ()
		win = gtk.GtkScrolledWindow ()
		win.set_policy (GTK.POLICY_NEVER, GTK.POLICY_AUTOMATIC)
		win.show ()
		# So we can still have the 'frame' look
		hbox.pack_start (win, padding=10)
		vbox.pack_start (hbox, padding=10)
		self.page.pack_start (vbox)

		# Create the scrolled viewport
		viewport = gtk.GtkViewport ()
		viewport.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		viewport.show ()
		win.add (viewport)

		# Create the packer for the thread objects
		self.thread_box = gtk.GtkPacker ()
		self.thread_box.show ()
		viewport.add (self.thread_box)

		# Add our idle counting function
		gtk.timeout_add (1000, self.idle_count)

	def set_max_threads (self, max):
		self.no_threads.set_value (max)

	def get_max_threads (self):
		return int (self.no_threads.get_value ())

	def update_transfer_totals (self, threads, speed):
		"""Updates the status of the current transfer speed for 'threads' number of threads.
		'speed' is in Kb."""

		self.transfer_label.set_text (_("%d Threads running, total transfer speed is %.02f Kb/s") %(threads, speed))

	def attach_main_thread (self, thread):
		"""Attachs the main thread object into the thread list."""

		self.main_thread = thread
		self.thread_box.add (thread.get_entry (), anchor=GTK.ANCHOR_N, options=GTK.FILL_X+GTK.FILL_Y)

	def use_main_thread (self, button):
		"""Updates the boolean flag indicated whether the main thread should be used for
		transfers instead of subsequent child threads."""

		# Set/unset the flag in one fell swoop
		self.main_transfer = (self.main_transfer + 1) %2

	def check_if_available_threads (self):
		"""Check if there are threads availablef or processing a queue entry (meaning they
		are idling or the thread count does not exceed the maximum). If a thread is
		available, assign it some work."""

		thread_entry = None

		server, file, size, attempt, direction, port, username, password, \
		remote_path, local_path, flags, method  = self.ctrlwin.queue_tab.get_next ()

		# Should we use the main thread or child threads?
		if self.main_transfer:
			if not self.main_thread.busy ():
				# Check if this file is on the same host. Otherwise, spawn a
				# transfer thread, since we don't really care about having multiple
				# connections to different hosts
				if self.main_thread.get_host () == server and \
				   self.main_thread.get_port () == int (port):
					thread_entry = self.main_thread
				else:
					thread_entry = self.create_thread (server, int (port), username, password)
		else:
			# First check if there's an idling thread
			found = 0

			for t in self.threads:
				if not t.busy ():
					found = 1
					# Check if we're still connected to the same host. If not, cancel
					# the thread and create a new one connected to the right host
					if t.get_host () == server and t.get_port () == int (port):
						thread_entry = t
					else:
						t.cancel_thread ()
						thread_entry = self.create_thread (server, int (port), username, password)
					break

			if not found:
				if len (self.threads) < self.no_threads.get_value ():
					thread_entry = self.create_thread (server, int (port), username, password)

		if thread_entry is not None:
			self.ctrlwin.queue_tab.remove_next ()
			cur_attempt, max_attempt = string.split (attempt, '/')

			if flags[0] == 'f':
				thread_entry.initiate_transfer (remote_path, local_path, file, int (size), int (direction), cur_attempt, int (method))
			elif flags[0] == 'd':
				thread_entry.recurse_directory (remote_path, local_path, file, int (size), int (direction), cur_attempt, int (method))
			else:
				thread_entry.recurse_link (remote_path, local_path, file, int (size), int (direction), cur_attempt, int (method))

	def create_thread (self, server, port, username, password):
		"""Creates a child transfer thread that connects to 'server' at 'port'
		with 'username':'password'."""

		# Use our app as default opts. Iterate because it's not
		# a true hash.
		opts = { }
		for key in main.app.keys ():
			opts[key] = main.app[key]

		opts['host'] = server
		opts['port'] = port
		opts['username'] = username
		opts['password'] = password

		thread_entry = threads.TransferThread (opts)
		self.add_thread_entry (thread_entry)
		return thread_entry

	def get_truesize (self, file, direction):
		"""Returns the size in bytes of 'file' from the appropriate file window,
		indicated by 'direction' which can be { ftp.DOWNLOAD, ftp.UPLOAD }."""

		if direction == ftp.DOWNLOAD:
			return main.app.remotewin.get_file_size (file)
		elif direction == ftp.UPLOAD:
			return main.app.localwin.get_file_size (file)
		return None

	def add_thread_entry (self, entry):
		"""Adds a thread entry to the thread list and displays its status."""

		self.threads.append (entry)
		self.thread_box.add (entry.get_entry (), anchor=GTK.ANCHOR_N, options=GTK.FILL_X+GTK.FILL_Y)

	def idle_count (self):
		"""Updates the idle count for threads that are not currently busy and
		checks to see if there are threads available to process queued items.
		Also updates the total transfer speed estimate."""

		total_rate = 0.0
		for t in self.threads:
			if not t.busy ():
				t.update_idle ()
				if t.finished ():
					self.thread_box.remove (t.get_entry ())
					self.threads.remove (t)
			elif t.isTransfering ():
				t.update_transfer_estimate ()
				total_rate = total_rate + t.transfer_speed ()

		# First check if there's remaining in the queue
		# If so, feed it into the threads
		if self.ctrlwin.queue_tab.length ():
			self.check_if_available_threads ()

		no_threads = len (self.threads)

		# Now check the main thread
		if self.main_thread is not None:
			# Count the main thread
			no_threads = no_threads + 1

			if self.main_thread.finished ():
				self.thread_box.remove (self.main_thread.get_entry ())
				# Set the thread references to None
				self.main_thread = None
				main.app.main_thread = None
			elif not self.main_thread.busy ():
				self.main_thread.update_idle ()
			elif self.main_thread.isTransfering ():
				self.main_thread.update_transfer_estimate ()
				total_rate = total_rate + self.main_thread.transfer_speed ()

		# Now update the totals
		self.update_transfer_totals (no_threads, total_rate)
		return gtk.TRUE

	def get_thread_list (self):
		"""Returns a list of currently active threads. If the main browsing
		thread is active, it is placed at the head of the list."""

		if self.main_thread is not None:
			thread_list = [ self.main_thread ]
			thread_list.extend (self.threads)
		else:
			thread_list = self.threads
		return thread_list

	def get_page (self):
		return self.page

class DownloadTab:

	"""The download tab object

	The tab's purpose is to display a list of past successful transfers. A spin
	button is provided indicating how many times a transfer should be attempted
	before giving up (in the event of error). Another button is provided to
	indicate the contents of a received archive."""

	status_headings = [ _('Filename'), _('Status'), _('Type'), _('Size'), _('Local Path') ]
	archive_headings = [ _('Filename'), _('Size') ]

	def __init__ (self, ctrl):
		"""Create the download tab and load the appropriate widgets. 'ctrl' is a
		reference to the parent container."""

		self.ctrlwin = ctrl

		self.page = gtk.GtkVBox ()
		self.page.show ()

		box = gtk.GtkHBox ()
		box.show ()
		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.set_space_size (10)
		toolbar.show ()
		box.pack_start (toolbar, padding=6)
		self.page.pack_start (box, expand=gtk.FALSE, padding=4)

		clear_pic, clear_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "clear.xpm"))
		archive_pic, archive_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "check_archive.xpm"))

		toolbar.append_item ("", _("Clear List"), _("Clear List"),
			gtk.GtkPixmap (clear_pic, clear_mask),
			self.clear_list)
		toolbar.append_item ("", _("Show Archive Details"), _("Show Archive Details"),
			gtk.GtkPixmap (archive_pic, archive_mask),
			self.show_archive)
		toolbar.append_space ()

		box = gtk.GtkHBox ()
		box.show ()
		self.max_attempts = gtk.GtkSpinButton (gtk.GtkAdjustment (0, 0, 16, 1, 10, 0), 1, 0)
		self.max_attempts.set_numeric (gtk.TRUE)
		self.max_attempts.show ()
		box.pack_start (self.max_attempts)
		label = gtk.GtkLabel (_("max attempts"))
		label.show ()
		box.pack_start (label, padding=4)
		toolbar.append_widget (box, "", "")

		win = gtk.GtkScrolledWindow ()
		win.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		win.show ()
		self.page.pack_start (win)

		self.status_list = gtk.GtkCList (len (self.status_headings), self.status_headings)
		self.status_list.connect ("select_row", self.select_status_items)
		for i in range (len (self.status_headings)):
			self.status_list.set_column_width (i, 75)
		# Explicitly make the filename column bigger
		self.status_list.set_column_width (0, 125)
		self.status_list.set_column_visibility (4, 0)
		self.status_list.set_selection_mode (GTK.SELECTION_MULTIPLE)
		self.status_list.show ()
		win.add_with_viewport (self.status_list)

		win = gtk.GtkScrolledWindow ()
		win.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		win.set_usize (-1, 200)
		win.show ()
		self.page.pack_start (win, expand=gtk.FALSE)

		self.archive_list = gtk.GtkCList (len (self.archive_headings), self.archive_headings)
		self.archive_list.connect ("select_row", self.select_archive_items)
		self.archive_list.set_column_width (0, 150)
		self.archive_list.set_selection_mode (GTK.SELECTION_MULTIPLE)
		self.archive_list.show ()
		win.add_with_viewport (self.archive_list)

	def set_max_attempts (self, attempts):
		self.max_attempts.set_value (attempts)

	def get_max_attempts (self):
		return int (self.max_attempts.get_value ())

	def clear_list (self, button):
		self.status_list.clear ()

	def show_archive (self, button):
		pass

	def select_status_items (self, list, r, c, event):
		pass

	def select_archive_items (self, list, r, c, event):
		pass

	def add_to_list (self, file):
		"""Add a downloaded file to the file list."""

		(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
		 st_atime, st_mtime, st_ctime) = os.stat (file)
		size = app.beautify_size (st_size)

		# Note that file is a full pathname so let's split it for the list
		path, filename = os.path.split (file)

		# For now, just put status and type to "OK" and "Unknown"
		# Will need to fix later
		self.status_list.append ( (filename, _("OK"), _("Unknown"), size, path) )

	def get_page (self):
		return self.page

class FailureTab:

	"""The failure tab object

	Displays a list of failure attempts during file transfers and their causes.
	The tab provides functionality that enables the user to resubmit failed jobs
	to the queue for another attempt."""

	headings = [ _('Server'), _('Filename'), _('Size'), _('Error Msg'), _('Direction'), _('Port'), _('Username'),
	             _('Password'), _('Remote Path'), _('Local Path'), _('Attempt'), _('Type'), _('Method') ]

	def __init__ (self, ctrl):
		"""Create the download tab and load the appropriate widgets. 'ctrl' is a
		reference to the parent container object."""

		self.ctrlwin = ctrl

		self.page = gtk.GtkVBox ()
		self.page.show ()

		box = gtk.GtkHBox ()
		box.show ()
		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.set_space_size (10)
		toolbar.show ()
		box.pack_start (toolbar, padding=6)
		self.page.pack_start (box, expand=gtk.FALSE, padding=4)

		clear_pic, clear_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "clear.xpm"))
		resubmit_pic, resubmit_mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "resubmit_job.xpm"))

		toolbar.append_item ("", _("Clear List"), _("Clear List"),
			gtk.GtkPixmap (clear_pic, clear_mask),
			self.clear_list)
		toolbar.append_item ("", _("Resubmit Job"), _("Resubmit Job"),
			gtk.GtkPixmap (resubmit_pic, resubmit_mask),
			self.resubmit_job)

		win = gtk.GtkScrolledWindow ()
		win.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		win.show ()
		self.page.pack_start (win)

		self.list = gtk.GtkCList (len (self.headings), self.headings)
		self.list.set_column_width (0, 80)
		self.list.set_column_width (1, 100)
		self.list.set_column_width (2, 80)
		self.list.set_column_width (3, 500)
		# Make the invisible, well, invisible
		for i in range (4, 13):
			self.list.set_column_visibility (i, 0)
		self.list.set_selection_mode (GTK.SELECTION_MULTIPLE)
		self.list.show ()
		win.add_with_viewport (self.list)

	def clear_list (self, button):
		self.list.clear ()

	def resubmit_job (self, button):
		"""Resubmits the selected job into the queue for another attempt at
		downloading. The attempt number is incremented appropriately."""

		max_attempts = self.ctrlwin.download_tab.get_max_attempts ()

		for i in self.list.selection:

			server, file, size, error, direction, port, username, password, \
			remote_path, local_path, attempt, type, method = self.get_list_entry (i)
			attempt = int (attempt)

			# Do nothing if we've exceeded max attempts
			if attempt > max_attempts:
				continue

			self.list.remove (i)
			self.ctrlwin.queue_tab.add_to_queue ({
				'file'        : file,
				'server'      : server,
				'port'        : port,
				'username'    : username,
				'password'    : password,
				'remote_path' : remote_path,
				'local_path'  : local_path,
				'size'        : size,
				'attempt'     : str (attempt + 1),
				'maxattempts' : max_attempts,
				'direction'   : direction,
				'flags'       : type,
				'method'      : method,
			})

	def get_list_entry (self, pos):
		"""Returns a tuple containing the data in the failure list at row
		'pos'."""

		entry = [ ]
		for i in range (self.list.columns):
			entry.append (self.list.get_text (pos, i))
		return tuple (entry)

	def add_to_list (self, hash):
		"""Adds an entry to the failure list, a transfer attempt whose attributes are
		specified in the dictionary 'hash'."""

		# Check if we've hit max attempts
		attempt = int (hash['attempt'])
		max_attempt = self.ctrlwin.download_tab.get_max_attempts ()
		if attempt > max_attempt:
			hash['error'] = _("Max Attempts Exceeded")

		self.list.append ((hash['server'], hash['file'], hash['size'], hash['error'], hash['direction'],
		                   hash['port'], hash['username'], hash['password'], hash['remote_path'],
		                   hash['local_path'], hash['attempt'], hash['type'], hash['method']))

	def get_page (self):
		return self.page



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import sys
import string # For backwards compatibility with 1.5.x
import socket, threading, re, os, errno, time
import exceptions
import main, app, logging

from intl import _

CRLF = '\r\n'

DOWNLOAD = 0
UPLOAD = 1

class FtpException (exceptions.Exception):

	"""FTP-specific exception

	This class marks exceptions which are raised during any of the FTP
	functionality."""

	def __init__ (self, args=None):
		self.args = args

class FtpControlConnection (threading.Thread):

	"""An FTP control session object

	This class provides an interface for managing an FTP control session. The
	control session handles all sending and receiving in a separate thread so
	that multiple threaded connections can be maintained simultaneously.

	The FtpControlConnection object contains synchornization primitives to
	manage sending and receiving of FTP messages, as all calls to the control
	interface come from separate threads of execution. All outgoing FTP
	commands are command objects and implement the abstract interface in the
	FTPAbstractCommand class. The control thread then manages the queuing
	of the outgoing objects and the receipt of their responses. Methods are
	also provided to allow some finer-grained control over the queueing
	process, viewing status, as well as retrieving the response data."""

	def __init__ (self, host, port, silence=0, logging=0):
		"""Creates a control connection object and initializes necessary
		instance variables. The 'host' and 'port' arguments indicate which
		the other party of the connection. The 'silence' argument indicates
		whether the status of the FTP commands should appear in the application
		command-window (not desirable for transfer threads) and 'logging'
		argument indicates whether or not sent and received FTP commands
		should be logged. A call to the connect method should follow directly
		after object creation."""

		threading.Thread.__init__ (self)
		self.setDaemon (1)

		# Initialize instance variables
		self.sock = None
		self.file = None
		self.multiline_code = None
		self.cmdobj = None
		self.exception_handler = None
		self.incoming = [ ]
		self.outgoing = [ ]
		self.outgoing_event = threading.Event ()
		self.die = 0
		self.host = host
		self.port = port
		self.silence = silence
		self.cmd_log = None
		self.set_log_active (logging)

	def connect (self):
		"""Sets up a connection to the FTP server in  a non-blocking fashion
		(to allow for aborting of the connection process). This creates the
		instance variable 'sock', which is a reference to the connection
		socket, and 'file', which is a file object representing the socket."""

		# Create the socket
		proto = socket.getprotobyname ('tcp')
		try:
			self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM, proto)
		except socket.error, strerror:
			raise FtpException (_("Error opening control socket: %s") %strerror)

		gtk.threads_enter ()
		if not self.silence: main.app.ftp_local_tell_user (_("Connecting to %s on port %d.") %(self.host, self.port))
		gtk.threads_leave ()

		self.sock.setblocking (0)

		try:
			err = self.sock.connect_ex ((self.host, self.port))
		except socket.error, strerror:
			# If we're in here, we probably caught an error such as hostname lookups
			# This is pretty fatal, so let's clear any exception handlers before raising
			# the exception
			self.clear_exception_handler ()
			raise FtpException (_("Error connecting to host %s: %s") %(self.host, strerror))

		gtk.threads_enter ()
		if not self.silence: main.app.ftp_local_tell_user (_("Connected to %s.") %self.host)
		gtk.threads_leave ()

		while err == errno.EINPROGRESS and not self.die:

			time.sleep (1)
			err = self.sock.getsockopt (socket.SOL_SOCKET, socket.SO_ERROR)

		if err != 0 or self.die:
			raise FtpException (_("Error connecting to host %s: %s") %(self.host, os.strerror (err)))
		else:
			self.sock.setblocking (1)
			self.file = self.sock.makefile ('rb')

	def destroy (self):
		"""Tear down the control connection gracefully."""

		# This might need some revising at some point
		if not self.silence: main.app.ftp_local_tell_user (_("Closing connection to %s...") %(self.host))
		self.die = 1
		self.outgoing_event.set ()
		if self.file is not None:
			self.file.close ()
		if self.sock is not None:
			self.sock.close ()
		if self.logging:
			self.log (_("Closing connection to %s:%d...") %(self.host, self.port))
			self.log ('')

	def get_host (self):
		"""Returns the host that the control connection is connected to."""

		return self.host

	def get_port (self):
		"""Returns the port that the control connection is using."""

		return self.port

	def reset_queue (self):
		"""Clears the outgoing FTP command queue and puts the connection thread in
		blocking mode."""

		self.outgoing = [ ]
		self.outgoing_event.set ()

	def wait_for_response (self):
		"""Provides a method for command objects to indicate that they would like
		to wait until the FTP server issues a response to their last command (usually
		always the case)."""

		while 1:
			line = self.file.readline ()

			if not line:
				gtk.threads_enter ()
				if not self.silence: main.app.ftp_local_tell_user (_("Connection to %s terminated...") %self.host)
				gtk.threads_leave ()
				self.sock.close ()
				return

			# Clean up the line
			if line[-2:] == CRLF: line = line[:-2]
			elif line[-1:] in CRLF: line = line[:-1]

			data = self.process_data (line)
			if data is not None:
				return data

	def process_data (self, line):
		"""Determines whether the response 'line' is part of a single-line response or
		a multi-line response. Also displays the response in the application command
		window if the control connection is not silent and logs the response line if
		appropriate. The response is stored in the instance variable 'self.incoming',
		to be retrieved if desired."""

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_remote_tell_user (line)
			gtk.threads_leave ()

		if self.logging:
			self.log (line)

		self.incoming.append (line)

		if line[3:4] == '-' or self.multiline_code is not None:
			if self.multiline_code is None:
				self.multiline_code = line[:3]
				return
			elif line[:3] == self.multiline_code and line[3:4] != '-':
				self.multiline_code = None
				data = self.incoming
				self.incoming = [ ]
				return data
		else:
			data = self.incoming
			self.incoming = [ ]
			return data

	def send (self, msg, sensitive=0):
		"""Appends an end of message marker to the FTP message string and sends
		the message."""

		if self.logging:
			if not sensitive:
				self.log (msg)
			else:
				# Star out the right-most word
				cmd, secret = string.split (msg, ' ')
				newsecret = ''
				for i in range (len (secret)):
					newsecret = newsecret + '*'
				self.log (string.join ((cmd, newsecret), ' '))
		return self.sock.send (msg + CRLF)

	def send_oob (self, msg):
		"""Appends an end of message marker to the FTP message string and
		toggles the out-of-band TCP flag when sending the message."""

		return self.sock.send (msg + CRLF, socket.MSG_OOB)

	def run (self):
		"""Executes the command objects from the outgoing queue in a separate
		execution thread. The thread blocks when the queue is empty. If an
		exception occurs during the execution of the command object, the
		exception handler is called if it is set or by default, an error message
		appears in the command window. The outgoing queue is also always reset
		with the advent of an exception."""

		while not self.die:
			if self.outgoing == [ ]:
				self.outgoing_event.clear ()
				self.outgoing_event.wait ()

			if not self.outgoing == [ ]:
				self.cmdobj = self.outgoing.pop (0)
				if self.cmdobj is None:
					continue
				else:
					try:
						# Call the execute function directly so
						# it runs in our thread
						self.cmdobj.execute ()
					except Exception, strerror:

						# Always reset the queue
						self.reset_queue ()

						if not self.die:
							if self.exception_handler is not None:
								self.exception_handler (strerror)
							else:
								gtk.threads_enter ()
								main.app.ftp_error_tell_user (strerror)
								gtk.threads_leave ()

	def set_exception_handler (self, callback):
		"""Sets an exception handler to deal with exceptions during command
		exceution asynchronously."""

		self.exception_handler = callback

	def get_executing_obj (self):
		"""Returns a reference to the last executed or currently executing
		command object. This is useful for interacting directly with the
		object after the command has been executed, such as retried the
		result of the response."""

		return self.cmdobj

	def busy (self):
		"""Indicates whether or not a command object is currently being
		executed."""

		return self.outgoing_event.isSet ()

	def add_outgoing (self, cmdobj):
		"""Adds the command object to the outgoing queue and unblocks the
		execution thread if necessary."""

		self.outgoing.append (cmdobj)
		if not self.outgoing_event.isSet ():
			self.outgoing_event.set ()

	def set_log_active (self, bool):
		"""Creates a logging object if 'bool' is 1, and destroy the
		logging object is bool is false."""

		if bool:
			if self.cmd_log is None:
				self.cmd_log = logging.Log (os.path.join (app.CONFIG_DIR, main.app['cmd_log']))
		else:
			if self.cmd_log is not None:
				del self.cmd_log
				self.cmd_log = None
		self.logging = bool

	def log (self, cmd):
		"""Writes the cmd string to the log."""

		self.cmd_log.write (cmd)

class FTPAbstractCommand:

	"""Abstract interface for FTP command objects."""

	def execute (self):
		"""This needs to be overriden in the derived class."""

		pass

class FTPGetWelcomeMsg (FTPAbstractCommand):

	"""Welcome Message command object

	A command object for receiving the welcome message. This command object
	should be added to the control queue shortly after establishing the
	connection."""

	def __init__ (self, queue):
		self.control_queue = queue

	def execute (self):

		response = self.control_queue.wait_for_response ()
		c = response[-1][:1]
		if c not in '123':
			raise FtpException (_("Error connecting to server: Welcome message not received correctly"))

class FTPConnect (FTPAbstractCommand):

	"""Connection command object

	Causes the control connection to initiate a connection with the FTP
	server in a separate execution thread."""

	def __init__ (self, queue):
		self.control_queue = queue

	def execute (self):
		self.control_queue.connect ()

class FTPAbortCmd (FTPAbstractCommand):

	"""Abort command object

	Sends an abort to the FTP server to abort the current message
	sequence."""

	def __init__ (self, queue, silence=0):

		self.control_queue = queue
		self.silence = silence

	def execute (self):

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user ("ABOR")
			gtk.threads_leave ()

		self.control_queue.send_oob ("ABOR")
		response = self.control_queue.wait_for_response ()
		if response[-1][:3] not in ('426', '226'):
			raise FtpException (response[-1])

class FTPVoidCmd (FTPAbstractCommand):

	"""General command object

	This models the basic FTP message exchange: a command is
	sent and execution pauses while a response is awaited. The
	VoidCmd object is used when we don't need any further processing
	on the response."""

	def __init__ (self, queue, cmd, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.cmd = cmd

	def execute (self):

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user (self.cmd)
			gtk.threads_leave ()

		self.control_queue.send (self.cmd)
		self.response = self.control_queue.wait_for_response ()
		if self.response[-1][0] != '2':
			raise FtpException (self.response[-1])

class FTPBasicCmd (FTPVoidCmd):

	"""Expanded command object

	Provides the same functionality as FTPVoidCmd, except provides
	a method for retrieving the actual response for further
	processing."""

	def __init__ (self, queue, cmd, silence=0):
		FTPVoidCmd.__init__ (self, queue, cmd, silence)

	def get_response (self):
		return self.response

class FTPPendingCmd (FTPAbstractCommand):

	"""Pending command object

	Pending commands are FTP commands which set up conditions
	for subsequent commands and thus should return a 3xx
	error code."""

	def __init__ (self, queue, cmd, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.cmd = cmd

	def execute (self):

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user (self.cmd)
			gtk.threads_leave ()

		self.control_queue.send (self.cmd)
		self.response = self.control_queue.wait_for_response ()
		if self.response[-1][0] != '3':
			raise FtpException (self.response[-1])


class FTPLogin (FTPAbstractCommand):

	"""Login command object

	Performs the necessary sequence to log the user into the
	FTP server with a 'user' name and password 'pw'. Acct is
	currently not used and ignored."""

	def __init__ (self, queue, user, pw, acct, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.user = user
		self.pw = pw
		self.acct = acct

	def execute (self):

		usr_string = "USER " + self.user

		if not self.silence:
			gtk.threads_enter ()
			main.app.ftp_local_tell_user (usr_string)
			gtk.threads_leave ()

		self.control_queue.send (usr_string)
		response = self.control_queue.wait_for_response ()

		if response[-1][0] == '3':

			pw_string = "PASS " + self.pw

			# Create a print friendly version
			chars = list (self.pw)
			for i in range (len (chars)):
				chars[i] = "*"
			safe_pw = "PASS " + string.join (chars, '')

			if not self.silence:
				gtk.threads_enter ()
				main.app.ftp_local_tell_user (safe_pw)
				gtk.threads_leave ()

			self.control_queue.send (pw_string, sensitive=1)
			response = self.control_queue.wait_for_response ()

		if response[-1][0] == '3':
			# For now, assume ACCT to be ''
			acct_string = "ACCT "

			if not self.silence:
				gtk.threads_enter ()
				main.app.ftp_local_tell_user (acct_string)
				gtk.threads_leave ()

			self.control_queue.send (acct_string)
			response = self.control_queue.wait_for_response ()

		if response[-1][0] != '2':
			raise FtpException (_("Error logging in to %s: %s") %(self.control_queue.get_host (), response[-1]))

class FTPTransferConnection:

	"""An FTP transfer object

	This class provides a wrapper interface for managing FTP transfers, which
	occur on different ports, separate from the control connection. Currently,
	only passive transfers (meaning the client initiates the connection to a
	port designated by the server and the only kind that works behind a
	firewall) are supported."""

	def __init__ (self, control, host, port, passive=1):
		"""Initializes the transfer object and establishes a connection to
		'host' on 'port'. The connection attempt is made in non-blocking mode
		to allow for aborting."""

		self.control_queue = control
		self.abort_flag = 0

		if passive:
			# Create the socket
			proto = socket.getprotobyname ('tcp')
			try:
				self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM, proto)
			except socket.error, strerror:
				raise FtpException (_("Error creating socket for transfer: %s") %strerror)

			self.sock.setblocking (0)

			try:
				err = self.sock.connect_ex ((host, port))
			except socket.error, strerror:
				raise FtpException (_("Error creating socket for transfer: %s") %(strerror))

			while err == errno.EINPROGRESS and not self.abort_flag:

				time.sleep (0.25)
				err = self.sock.getsockopt (socket.SOL_SOCKET, socket.SO_ERROR)

			if err != 0 or self.abort_flag:
				raise FtpException (_("Error creating socket for transfer: %s") %os.strerror (err))
			else:
				self.sock.setblocking (1)
		else:
			# This is currently untested
			host, port = self.create_server_sock ()
			try:
				self.sock = socket.socket (socket.AF_INET, socket.SOCK_STREAM)
				self.sock.bind (('', 0))
				self.sock.listen (1)
				dummyhost, port = self.sock.getsockname ()
				host, dummyport = control.getsockname ()
			except socket.error, (strerror):
				raise FtpException (_("Error creating socket for transfer: %s") %strerror)

			self.send_port (host, port)
			self.sock, sockaddr = self.sock.accept ()

	def send_port (self, host, port):
		"""Provides a method for sending the 'host'/'port' combination for
		a non-passive FTP transfer (meaning the listening port is opened
		on the client-side)."""

		hbytes = string.split (host, '.')
		pbytes = [ `port/256`, `port%256` ]
		bytes = hbytes + pbytes
		cmd = 'PORT ' + string.join (bytes, ',')
		return self.control.basic_cmd (cmd)

	def create_server_sock (self):
		pass

	def get_sock (self):
		return self.sock

	def get_file (self):
		return self.sock.makefile ('rb')

	def close (self):
		self.sock.close ()

	def abort (self):
		self.abort_flag = 1

class FTPTransferCmd (FTPAbstractCommand):

	"""The issue transfer sequence command object

	This models the message sequence necessary for initiating the
	actual transfer."""

	# For parsing server return codes
	re150 = re.compile ("150 .* \((\d+) bytes\)", re.IGNORECASE)

	def __init__ (self, queue, cmd, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.cmd = cmd
		self.size = None

	def execute (self):

		self.control_queue.send (self.cmd)
		response = self.control_queue.wait_for_response ()

		if response[-1][0] != '1':
			raise FtpException (_("Error initiating transfer: %s") %response[-1])
		if response[-1][:3] == '150':
			self.size = self.parse150 (response[-1])

	def parse150 (self, resp):
		"""Parses the 150 return code in response string 'resp' given after a
		download is initiated for retrieval of the file size."""

		# Sanity checking
		if resp[:3] != '150':
			raise FtpException (_("Error parsing 150 response: 150 response not given"))

		matched = self.re150.match (resp)
		if matched:
			return int (matched.group (1))
		return None

	def get_size (self):
		"""Returns the file size retrieved from the 150 return code while the
		transfer is initiated."""

		return self.size


class FTPTransfer:

	"""FTP transfer object

	This models the sequence necessary for initaited an FTP transfer of any
	type (textual/binary). It does not say anything about the type of transfer
	but rather provides an interface for settings up and managing the transfer
	connection."""

	def __init__ (self, queue, rest=None, passive=1, silence=0):

		self.control_queue = queue
		self.silence = silence
		self.rest = rest
		self.passive = passive

	def initiate_transfer (self):
		"""Creates and initiates the connection for file transfer (negotiating a
		transfer port and the actual connection process."""

		if self.passive:
			cmd = FTPBasicCmd (self.control_queue, 'PASV', silence=self.silence)
			cmd.execute ()
			host, port = self.parse227 ( cmd.get_response()[-1] )

			self.transfer = FTPTransferConnection (self.control_queue, host, port, passive=self.passive)

			if self.rest is not None:
				cmd = FTPPendingCmd (self.control_queue, "REST %s" %self.rest, silence=self.silence)
				cmd.execute ()

			if not self.silence:
				gtk.threads_enter ()
				main.app.ftp_local_tell_user (self.cmd)
				gtk.threads_leave ()

			cmd = FTPTransferCmd (self.control_queue, self.cmd, silence=self.silence)
			cmd.execute ()

			self.size = cmd.get_size ()
		else:
			# Non passive stuff will go here
			pass

	def parse227 (self, resp):
		"""Parses the response string 'resp' which contains the passive port to connect to
		when setting up the transfer connection."""

		# Sanity checking
		if resp[:3] != '227':
			raise FtpException (_("Error parsing 227 response: 227 response not given"))

		left_bracket = string.find (resp, '(')
		right_bracket = string.find (resp, ')')
		if left_bracket < 0 or right_bracket < 0:
			raise FtpException (_("Error parsing 227 response: Format should be (h1,h2,h3,h4,p1,p2) and given: %s") %resp)
		fields = string.split (resp[left_bracket + 1: right_bracket], ',')
		if len (fields) != 6:
			raise FtpException (_("Error parsing 227 response: Unable to pack appropriate number of fields"))
		host = string.join (fields[:4], '.')
		port = (int (fields[4]) << 8) + int (fields[5])
		return host, port

	def abort (self):
		self.transfer.abort ()

class FTPList (FTPTransfer):

	"""An FTP list command object

	Models the sequence necessary for retrieving a remote FTP listing. Methods
	are also provided for performing futher processing on the data."""

	def __init__ (self, queue, cmd, callback=None, rest=None, passive=1, silence=0):
		"""Initializes the list object by calling the base class to set up the
		actual transfer and initializes necessary instance variables. 'callback' is
		a function reference that is executed at the end of the transfer with a
		reference to the data. If 'callback' is none, the remote file window is
		updated by default."""

		FTPTransfer.__init__ (self, queue, rest=rest, passive=passive, silence=silence)
		self.control_queue = queue
		self.silence = silence
		self.callback = callback
		self.cmd = cmd
		self.abort_flag = 0
		self.data = [ ]              # Local copy of data

	def execute (self):

		# Set the appropriate transfer type
		type = FTPBasicCmd (self.control_queue, 'TYPE A', silence=self.silence)
		type.execute ()

		# Now prepare for the transfer
		self.initiate_transfer ()

		self.sock = self.transfer.get_sock ()
		self.file = self.transfer.get_file ()

		bytes = 0

		while not self.abort_flag:
			line = self.file.readline ()
			if not line:
				break
			if line[-2:] == CRLF:
				line = line[:-2]
			elif line[-1:] in CRLF:
				line = line[:-1]

			# For updating the status window
			bytes = bytes + len (line)
			if not self.silence:
				gtk.threads_enter ()
				main.app.update_list_transfer (bytes)
				gtk.threads_leave ()

			self.data.append (line)

		# Wait for the finished transfer and then get the response
		self.file.close ()
		self.sock.close ()
		self.control_queue.wait_for_response ()

		# Now reset the status window to the current directory
		gtk.threads_enter ()
		main.app.list_transfer_finished ()
		gtk.threads_leave ()

		# Check if we should execute the call back
		if self.callback is None:
			gtk.threads_enter ()
			main.app.update_remote_listing (self.get_data ())
			gtk.threads_leave ()
		else:
			self.callback (self.get_data ())

	def get_data (self):
		"""Returns a reference to an array of the actual listing."""

		return self.data

	def abort (self):

		self.transfer.abort ()
		self.abort_flag = 1

class FTPAsciiTransfer (FTPTransfer):

	"""An FTP ascii transfer command object

	Models the sequence necessary for performing an ascii file transfer and
	handles writing the transferred file to disk."""

	def __init__ (self, queue, cmd, local_file, direction, progress=None, passive=1, silence=0):
		"""Initializes the ascii transfer object by first setting up the transfer
		connection when the base class is initialized and then appropriately
		initalizing instance variables. The 'direction' variable can take on the
		values of ftp.DOWNLOAD and ftp.UPLOAD respectively. 'progress' points to
		an object that can implements the "update_transfer" method and allows for
		tracking of the transfer's status. A file is created from the socket and
		readline is used for the actual transfer."""

		FTPTransfer.__init__ (self, queue, passive=passive)
		self.control_queue = queue
		self.cmd = cmd
		self.local_file = local_file
		self.direction = direction
		self.silence = silence
		self.progress = progress
		self.passive = passive
		self.abort_flag = 0               # For aborting the transfer

	def execute (self):

		# Set the appropriate transfer type
		type = FTPBasicCmd (self.control_queue, 'TYPE A', silence=self.silence)
		type.execute ()

		# Now prepare for the transfer
		self.initiate_transfer ()

		self.sock = self.transfer.get_sock ()
		self.file = self.transfer.get_file ()

		if self.direction == DOWNLOAD:
			file = open (self.local_file, 'w')

			while not self.abort_flag:
				line = self.file.readline ()
				if not line:
					break
				if self.progress is not None:
					gtk.threads_enter ()
					self.progress.update_transfer (len (line))
					gtk.threads_leave ()
				file.write (line)
		elif self.direction == UPLOAD:
			file = open (self.local_file, 'r')

			while not self.abort_flag:
				line = file.readline ()
				if not line:
					break
				self.sock.send (line)
				if self.progress is not None:
					gtk.threads_enter ()
					self.progress.update_transfer (len (line))
					gtk.threads_leave ()

		# Wait for the finished transfer and then get the response
		self.file.close ()
		self.sock.close ()
		self.control_queue.wait_for_response ()

		# Update our status
		if self.direction == DOWNLOAD:
			gtk.threads_enter ()
			main.app.add_to_download_list (self.local_file)
			main.app.update_local_listing ('.')
			gtk.threads_leave ()
		elif self.direction == UPLOAD:
			gtk.threads_enter ()
			main.app.main_thread.list ()
			gtk.threads_leave ()

	def abort (self):
		self.abort_flag = 1

class FTPBinaryTransfer (FTPTransfer):

	"""An FTP binary transfer command object

	Models the sequence necessary for performing a binary file transfer and
	handles writing the transferred file to disk."""

	def __init__ (self, queue, cmd, local_file, direction, progress=None, blocksize=8192, rest=None, passive=1, silence=0):
		"""Initializes the binary transfer object by first setting up the transfer
		connection when the base class is initialized and then appropriately
		initalizing instance variables. The 'direction' variable can take on the
		values of ftp.DOWNLOAD and ftp.UPLOAD respectively. 'progress' points to
		an object that can implements the "update_transfer" method and allows for
		tracking of the transfer's status. The 'blocksize' is used for figuring
		out how much should be read from the socket and written to disk in a
		given time."""

		FTPTransfer.__init__ (self, queue, rest, passive)
		self.control_queue = queue
		self.cmd = cmd
		self.local_file = local_file
		self.direction = direction
		self.silence = silence
		self.progress = progress
		self.blocksize = blocksize
		self.rest = rest
		self.passive = passive
		self.abort_flag = 0               # For aborting the transfer

	def execute (self):

		# Set the appropriate transfer type
		type = FTPBasicCmd (self.control_queue, 'TYPE I', silence=self.silence)
		type.execute ()

		# Now prepare for the transfer
		self.initiate_transfer ()

		self.sock = self.transfer.get_sock ()

		# Perform the actual transferring of the data
		if self.direction == DOWNLOAD:
			if self.rest is not None:
				file = open (self.local_file, 'ab')
			else:
				file = open (self.local_file, 'wb')

			while not self.abort_flag:
				data = self.sock.recv (self.blocksize)
				if not data:
					break
				if self.progress is not None:
					gtk.threads_enter ()
					self.progress.update_transfer (len (data))
					gtk.threads_leave ()
				file.write (data)
		elif self.direction == UPLOAD:
			file = open (self.local_file, 'rb')

			if self.rest is not None:
				file.seek (self.rest)

			while not self.abort_flag:
				data = file.read (self.blocksize)
				if not data:
					break
				self.sock.send (data)
				if self.progress is not None:
					gtk.threads_enter ()
					self.progress.update_transfer (len (data))
					gtk.threads_leave ()

		file.close ()
		self.sock.close ()
		self.control_queue.wait_for_response ()

		# Update our status
		if self.direction == DOWNLOAD:
			gtk.threads_enter ()
			main.app.add_to_download_list (self.local_file)
			main.app.update_local_listing ('.')
			gtk.threads_leave ()
		elif self.direction == UPLOAD:
			gtk.threads_enter ()
			main.app.main_thread.list ()
			gtk.threads_leave ()

	def abort (self):
		self.abort_flag = 1

class FTP:

	"""The FTP core functionality class

	This class provides a wrapper for managing all aspects of the ftp session,
	including settings up, tearing down, and issuing commands in the ftp
	control session, as well as initiating file transfers. This class provides
	the all-encompassing interface to FTP functionality. Each instance is useful
	for interacting with a single FTP session."""

	def __init__ (self, host=None, port=None, silence=0, logging=0):
		"""Some sanity checking is performed and the control session object
		is created and initialized with destination 'host':'port'."""

		# Initialize instance variables
		self.passive_server = 1
		self.silence = silence
		self.handler = None

		# Sanity checking on our options
		if host is None:
			raise FtpException (_("Error opening FTP connection: No host given"))
		if port is None or not (port > 0 and port < 65534):
			raise FtpException (_("Error opening FTP connection: Invalid port given"))

		self.control_connect = FtpControlConnection (host, port, silence=self.silence, logging=logging)
		self.control_connect.start ()

	def destroy (self):
		"""Destroys the current control session, tearing down the connection."""

		self.control_connect.destroy ()

	def set_exception_handler (self, callback):
		"""Sets an exception handler of type:
		      def handler (msg)
		which is executed upon receipt of an exception during execution in a
		separate thread. Setting a handler of None clears the exception
		handler"""

		self.handler = callback

	def set_log_active (self, bool):
		"""Toggles whether logging of FTP commands and transfers is active."""

		self.control_connect.set_log_active (bool)

	def reset (self):
		"""Clears the outgoing command queue."""

		self.control_connect.reset_queue ()

	def ftp_busy (self):
		"""Indicates whether the ftp control session is currently busy sending
		and receiving FTP message."""

		return self.control_connect.busy ()

	def add_outgoing (self, obj):
		"""Sets up the current exception handler and adds the object to
		the control object's outgoing queue."""

		self.control_connect.set_exception_handler (self.handler)
		self.control_connect.add_outgoing (obj)

	def initiate_connect (self):
		"""Sets up the FTP control session connection."""

		connect = FTPConnect (self.control_connect)
		self.add_outgoing (connect)

	def get_welcome_msg (self):
		"""Receives the FTP welcome message, which should be done promptly after
		establishing a connection."""

		welcome = FTPGetWelcomeMsg (self.control_connect)
		self.add_outgoing (welcome)

	def perform_login (self, username, password):
		"""Performs the login sequence using 'username' and 'password'."""

		login = FTPLogin (self.control_connect, username, password, '', silence=self.silence)
		self.add_outgoing (login)

	def noop (self):
		"""Performs a "no operation". Useful for keeping the FTP connection
		from idling."""

		cmd = 'NOOP'
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)

	def cwd (self, dir):
		"""Changes the remote directory to 'dir'."""

		# Default to '/' if dir is blank
		if not dir:
			dir = '/'

		if dir == '..':
			cmd = "CDUP"
		elif dir == ".":
			# Actually do nothing and return
			return
		else:
			cmd = "CWD " + dir
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)

	def delete (self, file):
		"""Deletes the remote 'file' in the current remote directory."""

		cmd = 'DELE ' + file
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)

	def rename (self, name, new_name):
		"""Renames the remote file from 'name' to 'new_name'."""

		cmd = 'RNFR ' + name
		ftpcmd = FTPPendingCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)
		cmd = 'RNTO ' + new_name
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)

	def mkdir (self, directory):
		"""Creates the 'directory' in the current remote directory."""

		cmd = 'MKD ' + directory
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)

	def rmdir (self, directory):
		"""Removes the 'directory' in the current remote directory."""

		cmd = 'RMD ' + directory
		ftpcmd = FTPVoidCmd (self.control_connect, cmd, silence=self.silence)
		self.add_outgoing (ftpcmd)

	def abort (self):
		"""Aborts the currently issuing command."""

		self.control_connect.reset_queue ()
		aborcmd = FTPAbortCmd (self.control_connect, silence=self.silence)

		# Check if we're cancelling an outstanding transfer such as a binary
		# transfer or list
		executing = self.control_connect.get_executing_obj ()
		if isinstance (executing, FTPTransfer):
			executing.abort ()
		elif isinstance (executing, FTPConnect):
			self.control_connect.destroy ()
		self.add_outgoing (aborcmd)

	def list (self, args=None, callback=None, rest=None):
		"""Retrieves a remote listing."""

		cmd = 'LIST'
		if args is not None:
			for i in args:
				if i:
					cmd = cmd + (' ' + i)
		listcmd = FTPList (self.control_connect, cmd, callback=callback, silence=self.silence)
		self.add_outgoing (listcmd)

	def retrieve_ascii (self, file, local_dir, progress, passive=1):
		"""Performs an ascii retrieval of a file."""

		cmd = 'RETR ' + file
		local_file = os.path.join (local_dir, file)
		asciicmd = FTPAsciiTransfer (self.control_connect, cmd, local_file, DOWNLOAD, progress, passive, silence=self.silence)
		self.add_outgoing (asciicmd)

	def upload_ascii (self, file, local_dir, progress, passive=1):
		"""Performs an ascii upload."""

		cmd = 'STOR ' + file
		local_file = os.path.join (local_dir, file)
		asciicmd = FTPAsciiTransfer (self.control_connect, cmd, local_file, UPLOAD, progress, passive, silence=self.silence)
		self.add_outgoing (asciicmd)

	def retrieve_binary (self, file, local_dir, progress, blocksize=8192, passive=1, rest=None):
		"""Performs a binary retrieval of a file."""

		cmd = 'RETR ' + file
		local_file = os.path.join (local_dir, file)
		binarycmd = FTPBinaryTransfer (self.control_connect, cmd, local_file, DOWNLOAD, progress, blocksize, rest, passive, silence=self.silence)
		self.add_outgoing (binarycmd)

	def upload_binary (self, file, local_dir, progress, blocksize=8192, passive=1, rest=None):
		"""Uploads a binary file."""

		cmd = 'STOR ' + file
		local_file = os.path.join (local_dir, file)
		binarycmd = FTPBinaryTransfer (self.control_connect, cmd, local_file, UPLOAD, progress, blocksize, rest, passive, silence=self.silence)
		self.add_outgoing (binarycmd)



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GTK
import os
import string # For backwards compatibility with 1.5.x
import main

from intl import _

class BrowseWin:

	"""Directory browsing object base class

	Contains the member functions that make up the base class for
	displaying a browsing window that allows for both local and
	remote changing of directories. The display offers an entry
	both for direct entry, as well as a tree for selecting the
	directory.
	"""

	def __init__ (self):
		"""Creates the browsing window instance, draws the widgets,
		and sets itself as model. Not that the base class does not
		cause self.dialog to become visible but rather relies on the
		sub class to do this, as it is best to make the dialog
		visible at the very end of intialization."""

		self.dialog = gtk.GtkDialog ()
		self.dialog.set_title (_("ftpcube - Change Directory"))
		self.dialog.set_policy (gtk.FALSE, gtk.FALSE, gtk.TRUE)
		self.dialog.set_position (gtk.WIN_POS_CENTER)
		self.dialog.set_usize (400, 350)
		self.dialog.connect ("key_press_event", self.key_press)
		icon_pic, icon_mask = gtk.create_pixmap_from_xpm (self.dialog, None, os.path.join (main.ICONS_PREFIX, "ftpcube.xpm"))
		self.dialog.set_icon (gtk.GtkPixmap (icon_pic, icon_mask))

		# Make ourselves modal
		gtk.grab_add (self.dialog)

		box = gtk.GtkHBox (spacing=75)
		box.show ()
		self.dialog.action_area.add (box, padding=95)

		ok_button = gtk.GtkButton (_("Ok"))
		ok_button.set_usize (50, 25)
		ok_button.connect ("clicked", self.ok)
		box.pack_start (ok_button, expand=gtk.FALSE)
		ok_button.show ()

		button = gtk.GtkButton (_("Cancel"))
		button.set_usize (50, 25)
		button.connect ("clicked", self.cancel)
		box.pack_start (button, expand=gtk.FALSE)
		button.show ()

		packer = gtk.GtkPacker ()
		packer.show ()
		self.dialog.vbox.add (packer)

		label = gtk.GtkLabel (_("Directory:"))
		label.show ()
		packer.add (label, anchor=GTK.ANCHOR_NW, i_pad_x=40, i_pad_y=10)

		self.entry_box = gtk.GtkEntry ()
		self.entry_box.show ()
		packer.add (self.entry_box, options=GTK.FILL_X, pad_x=35)

		frame = gtk.GtkFrame ()
		frame.show ()
		packer.add (frame, options=GTK.EXPAND+GTK.FILL_X+GTK.FILL_Y, pad_x=15, pad_y=15)
		scrolled = gtk.GtkScrolledWindow ()
		scrolled.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		scrolled.show ()
		frame.add (scrolled)
		self.tree = gtk.GtkCTree ()
		self.tree.connect ("select_row", self.tree_select)
		self.tree.show ()
		scrolled.add (self.tree)

		self.load_pixmaps ()

	def load_pixmaps (self):
		"""Loads pixmaps for the browse tree."""

		self.folder_pic, self.folder_mask = gtk.create_pixmap_from_xpm (self.tree, None, os.path.join (main.ICONS_PREFIX, "folder.xpm"))

	def key_press (self, win, event):
		"""Looks for carriage return key pressed and invokes the ok handler
		if found."""

		if hasattr (event, 'string') and event.string == '\r':
			self.ok (None)

	def ok (self, button):
		"""A basic handler for the ok action button."""

		self.cancel (button)

	def cancel (self, button):
		"""A basic handler for the cancel action button."""

		self.dialog.hide ()
		self.dialog.destroy ()

	def tree_select (self, tree, row, col, event):
		"""Updates the text entry box when an entry in the tree is selected."""

		if hasattr (event, 'type'):
			node = self.tree.node_nth (row)
			path = self.node_to_path (node)
			self.entry_box.set_text (path)

	def node_to_path (self, node):
		"""To Be overridden in the subclass. Finds a node in the hierachy
		and returns its path."""

		pass

class LocalCWD (BrowseWin):

	"""Local Change Directory Window Object

	Displays a window allowing the user to change the local directory.
	The tree begins expanded to the user's current directory. If the
	user clicks on a directory that itself contains further
	sub-directories, these sub-directories are automatically expanded
	and added to the tree."""

	def __init__ (self, path):
		"""Create the window instance, load the contents of the tree
		and expand to our current path."""

		BrowseWin.__init__ (self)
		self.tree.show_stub = 1

		# Load in the list contents
		insert_node = self.tree.insert_node (None, None, [ os.sep ],
			pixmap_closed = self.folder_pic,
			mask_closed = self.folder_mask,
			pixmap_opened = self.folder_pic,
			mask_opened = self.folder_mask,
			is_leaf = gtk.FALSE)
		self.tree.toggle_expansion (insert_node)
		self.load_tree (path, insert_node)

		# Set the entry_box
		self.entry_box.set_text (path)

		self.dialog.show ()

	def ok (self, button):
		"""Performs the actual changing of local directory upon the
		ok action."""

		BrowseWin.ok (self, button)
		path = self.entry_box.get_text ()
		main.app.update_local_listing (path)

	def load_tree (self, path, parent):
		"""Loads the hierachy along 'path' using 'parent' as its
		root. Only directories are added as sub-directories during
		expansion."""

		dirs = [ ]
		full_path = os.sep
		for item in string.split (path, os.sep):
			if item:
				full_path = os.path.join (full_path, item)
				if os.path.isdir (full_path):
					dirs.append (item)
		dirs.insert (0, os.sep)

		cur_path = os.sep
		while dirs:
			dir = dirs.pop (0)
			cur_path = os.path.join (cur_path, dir)
			listing = self.get_dir_listing_only (cur_path)

			# Insert the new node and assign the parent to the appropriate node
			new_parent = None
			for item in listing:
				node = self.tree.insert_node (parent, None, [ item ],
					pixmap_closed = self.folder_pic,
					mask_closed = self.folder_mask,
					pixmap_opened = self.folder_pic,
					mask_opened = self.folder_mask,
					is_leaf = gtk.FALSE)
				self.tree.toggle_expansion (node)
				if dirs and item == dirs[0]:
					new_parent = node
			parent = new_parent

	def get_dir_listing_only (self, path):
		"""Converts the directory listing at 'path' to a listing of
		only sub-directories. A blank list is returned if an exception
		occurred while generating the listing."""

		try:
			listing = os.listdir (path)
		except OSError, strerror:
			return [ ]
		# Make sure we're only looking at directories
		listing = filter (lambda x, path=path: os.path.isdir (os.path.join (path, x)), listing)
		return listing

	def tree_select (self, tree, row, col, event):
		"""Upon selection of a directory, the entry box is updated with
		the path of the currently selected directory. If any sub-directories
		exist, they are expanded."""

		if hasattr (event, 'type'):
			
			node = self.tree.node_nth (row)
			path = self.node_to_path (node)
			self.entry_box.set_text (path)

			if not node.children:
				self.expand_node (node, path)
				self.tree.expand (node)

	def expand_node (self, parent, path):
		"""This expands a single node by adding 'parent's sub-directories
		into the tree with the given 'path'."""

		listing = self.get_dir_listing_only (path)
		self.tree.freeze ()
		for item in listing:
			node = self.tree.insert_node (parent, None, [ item ],
				pixmap_closed = self.folder_pic,
				mask_closed = self.folder_mask,
				pixmap_opened = self.folder_pic,
				mask_opened = self.folder_mask,
				is_leaf = gtk.FALSE)
		self.tree.thaw ()

	def node_to_path (self, node):
		"""Converts a single node to its hierachical path representation."""

		path = [ ]
		while node is not None:
			path.insert (0, self.tree.get_node_info (node)[0])
			node = node.parent
		path = path[0] + string.join (path[1:], os.sep)
		return path

class RemoteCWD (BrowseWin):

	"""Remote Change Directory Window Object

	Displays a window allowing the user to change the remote directory.
	The tree begins expanded to the user's current directory. This window
	makes use of the remote file window caching object to load the
	remote listing. At any given time, the tree spans the entire contents
	of the directory cache and only shows the fringes of the directories
	that the user has visited."""

	def __init__ (self, cache, path):
		"""Creates the window instance and loads the cache into the tree.
		The tree is then expanded along path so that the user starts in
		his current remote directory."""

		BrowseWin.__init__ (self)
		self.cache = cache

		# Load in the list contents
		insert_node = self.tree.insert_node (None, None, [ self.cache.get_delim () ],
			pixmap_closed = self.folder_pic,
			mask_closed = self.folder_mask,
			pixmap_opened = self.folder_pic,
			mask_opened = self.folder_mask,
			is_leaf = gtk.FALSE)
		self.load_tree (self.cache.get_root (), insert_node)

		# Set ourselves to inital contents
		self.expand_path (path)
		self.entry_box.set_text (path)

		self.dialog.show ()

	def ok (self, button):
		"""Executes the remote change of directory upon clicking of the
		ok action button."""

		BrowseWin.ok (self, button)
		path = self.entry_box.get_text ()
		main.app.main_thread.cwd (path)

		# Messy but will have to rework these later
		main.app.remotewin.set_dir (path)
		main.app.remotewin.get_listing ()

	def load_tree (self, node, parent):
		"""Loads the directory structure from within the cache into the tree. Only
		directories are put into the tree."""

		directories = filter (lambda x: not x[0] == '..' and x[3][0] == 'd', node.data)
		for dir in directories:
			dir_name = dir[0]
			insert_node = self.tree.insert_node (parent, None, [ dir_name ],
				pixmap_closed = self.folder_pic,
				mask_closed = self.folder_mask,
				pixmap_opened = self.folder_pic,
				mask_opened = self.folder_mask,
				is_leaf = gtk.FALSE)
			child = self.cache.get_child_node (node, dir_name)
			if child is not None:
				self.load_tree (child, insert_node)

	def node_to_path (self, node):
		"""Converts a node to its hierachical path representation."""

		path = [ ]
		while node is not None:
			path.insert (0, self.tree.get_node_info (node)[0])
			node = node.parent
		path = path[0] + string.join (path[1:], self.cache.get_delim ())
		return path

	def expand_path (self, path):
		"""Expands the nodes along 'path' in the tree. This is used primary
		for expanded the tree to the user's current remote directory."""

		path = filter (lambda x: x, string.splitfields (path, self.cache.get_delim ()))
		node = self.tree.node_nth (0)
		self.tree.expand (node)
		if not path:
			return

		while path:
			node_name = path.pop (0)
			child_names = map (lambda x, tree=self.tree: tree.get_node_info (x)[0], node.children)
			try:
				child_index = child_names.index (node_name)
			except ValueError, strerror:
				return None
			node = node.children[child_index]
			self.tree.expand (node)



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import sys, os, time, re
import string # For backwards compatibility with 1.5.x
import main, app, ftp, cache
import browsewin

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class AbstractFileWin:

	UNSORTED = 0
	SORT_BY_NAME = 1
	SORT_BY_SIZE = 2
	SORT_BY_DATE = 3

	hidden_re = re.compile ("^\..*")

	def __init__ (self, container, headings):

		# Initialize instance variables
		self.sorted = 1
		self.directories_first = 1
		self.hidden_files = 1
		self.dir = ''

		box = gtk.GtkVBox ()
		box.show ()
		container.add (box)

		self.win = gtk.GtkScrolledWindow ()
		self.win.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		self.win.show ()
		box.pack_start (self.win)

		self.status_bar = gtk.GtkHBox ()
		self.status_bar.set_usize (-1, 15)
		self.status_bar.show ()
		box.pack_start (self.status_bar, expand=gtk.FALSE, fill=gtk.TRUE)

		self.list = gtk.GtkCList (len (headings), headings)
		self.list.connect ("select_row", self.select_items)
		self.list.connect ("button_press_event", self.popup_menu)

		self.list.set_selection_mode (GTK.SELECTION_MULTIPLE)
		self.win.add_with_viewport (self.list)
		self.list.show ()

		self.load_pixmaps ()

	def load_pixmaps (self):

		self.folder_pixmap, self.folder_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "folder.xpm"))
		self.link_pixmap, self.link_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "link.xpm"))
		self.file_pixmap, self.file_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "file.xpm"))

	def select_pixmap (self, listitem):

		file, size, mtime, type, truesize = listitem
		if not type:
			return self.folder_pixmap, self.folder_mask
		else:
			if type[0] == 'l':
				return self.link_pixmap, self.link_mask
			elif type[0] in 'd ':
				return self.folder_pixmap, self.folder_mask
			else:
				return self.file_pixmap, self.file_mask

	def select_items (self, list, r, c, event):

		# This needs to be overridden in the subclass
		pass

	def popup_menu (self, list, event):

		# This needs to be overridden in the subclass
		pass

	def hide_hidden (self, button):

		self.hidden_files = (self.hidden_files + 1) % 2
		self.update_listing ('.')

	def unsorted (self, action, button):

		self.sorted = self.UNSORTED
		self.directories_first = 0
		self.update_listing ('.')

	def sort_by_name (self, action, button):

		self.sorted = self.SORT_BY_NAME
		self.directories_first = 1
		self.update_listing ('.')

	def sort_by_size (self, action, button):

		self.sorted = self.SORT_BY_SIZE
		self.directories_first = 1
		self.update_listing ('.')

	def sort_by_date (self, action, button):

		self.sorted = self.SORT_BY_DATE
		self.directories_first = 1
		self.update_listing ('.')

	def sort_directories_first (self, button):

		self.directories_first = (self.directories_first + 1) % 2
		self.update_listing ('.')

	def clear_items (self):
		self.list.clear ()

	def get_dir (self):
		return self.dir

	def set_dir (self, dir):
		self.dir = dir

	def get_file_size (self, file):

		for i in range (self.list.rows):
			if self.list.get_pixtext (i, 0)[0] == file:
				return int (self.list.get_text (i, 4))
		return None

	def update_listing (self, dir):

		# To be overridden in the subclass
		raise NotImplementedError

	def sort_list (self, list):

		# To be overridden in the subclass
		raise NotImplementedError

	def perform_sort (self, list):

		if self.sorted == self.SORT_BY_NAME:
			list.sort (lambda x, y: cmp (x[0], y[0]))
		elif self.sorted == self.SORT_BY_SIZE:
			list.sort (lambda x, y: cmp (int (x[4]), int (y[4])))
		elif self.sorted == self.SORT_BY_DATE:
			list.sort (lambda x, y: cmp (time.strptime (x[2], "%b %d %H:%M"), time.strptime (y[2], "%b %d %H:%M")))
		return list

class LocalWin (AbstractFileWin):

	headings = [ _('Filename'), _('Size'), _('Date'), _('Type'), _('Truesize') ]
	time_re = re.compile ('[A-Za-z]+\s+([A-Za-z]+\s+\d+\s+\d+:\d+):(\d+)\s+\d+')

	def __init__ (self, container):

		AbstractFileWin.__init__ (self, container, self.headings)

		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.show ()
		pic, mask = gtk.create_pixmap_from_xpm (toolbar, None, os.path.join (main.ICONS_PREFIX, "enable.xpm"))
		toolbar.append_element (GTK.TOOLBAR_CHILD_TOGGLEBUTTON, None,
			"", _("Enable/Disable automatic updates"),
			_("Enable/Disable automatic updates"),
			gtk.GtkPixmap (pic, mask),
			None)
		self.status_bar.pack_start (toolbar, expand=gtk.FALSE, fill=gtk.FALSE)

		self.status_label = gtk.GtkLabel ("")
		self.status_label.show ()
		self.status_bar.pack_start (self.status_label, expand=gtk.FALSE, fill=gtk.FALSE, padding=6)

		# Set the settings for our list
		for i in range(len (self.headings)):
			self.list.set_column_width (i, 100)
			self.list.set_column_justification (i, GTK.JUSTIFY_LEFT)
		# Explicitly make the first column for the filename bigger
		self.list.set_column_width (0, 200)
		# Hide the flags column
		self.list.set_column_visibility (3, 0)
		self.list.set_column_visibility (4, 0)
		# Make file sizes right justified
		self.list.set_column_justification (1, GTK.JUSTIFY_RIGHT)
		# So we can't click the titles as buttons
		self.list.column_titles_passive ()

		# Open up the default directory for display
		if os.environ.has_key ('HOME'):
			self.dir = os.environ['HOME']
		else:
			self.dir = os.getcwd ()
		self.update_listing (self.dir)

	def select_items (self, list, r, c, event):

		if hasattr (event, 'type'):
			# Check for double clicks
			if event.type == 5:
				selected = list.get_pixtext (r, 0)[0]
				self.update_listing (selected)

	def popup_menu (self, list, event):

		if event.button == 3:
			opts = [
				(_('/Upload Files'),             None, self.upload_selected,  0, ''),
				#('/Open',                     None, self.open_selected,    0, ''),
				#('/Edit',                     None, self.edit_selected,    0, ''),
				#('/View File',                None, None,                  0, ''),
				(_('/Rename Files'),             None, self.rename_selected,  0, ''),
				(_('/Delete Selected Files'),    None, self.delete_selected,  0, ''),
				('/Sep1',                     None, None,                  0, '<Separator>'),
				(_('/Create Directory'),         None, self.create_directory, 0, ''),
				('/Change Directory',            None, self.cd_selected,      0, ''),
				#('/Change Drive',             None, None,                  0, ''),
				#('/Sep2',                     None, None,                  0, '<Separator>'),
				#('/Compare Directories',      None, None,                  0, ''),
				#('/Open Current in Browser',  None, None,                  0, ''),
				#('/Open Selected in Browser', None, None,                  0, ''),
				#('/Command Prompt Here',      None, None,                  0, ''),
				#('/Sep3',                     None, None,                  0, '<Separator>'),
				#('/Directory Info',           None, None,                  0, ''),
				('/Sep4',                     None, None,                  0, '<Separator>'),
				(_('/Hide Hidden Files'),        None, None,                  0, '<ToggleItem>'),
				('/Sep5',                     None, None,                  0, '<Separator>'),
				(_('/Sort Order'),               None, None,                  0, '<Branch>'),
				(_('/Sort Order/Unsorted'),      None, self.unsorted,         0, ''),
				(_('/Sort Order/Sort by Name'),  None, self.sort_by_name,     0, ''),
				(_('/Sort Order/Sort by Size'),  None, self.sort_by_size,     0, ''),
				(_('/Sort Order/Sort by Date'),  None, self.sort_by_date,     0, ''),
				(_('/Sort Order/<separator>'),   None, None,                  0, '<Separator>'),
				(_('/Sort Order/Order Directories-Links-Files'), None, None, 0, '<ToggleItem>')
			]

			ag = gtk.GtkAccelGroup()
			items = gtk.GtkItemFactory(gtk.GtkMenu.get_type(), "<popup>", ag)
			items.create_items (opts)

			checked = items.get_widget (_('/Hide Hidden Files'))
			if self.hidden_files:
				checked.set_active (gtk.TRUE)
			checked.connect ("activate", self.hide_hidden)

			checked = items.get_widget (_('/Sort Order/Order Directories-Links-Files'))
			if self.directories_first:
				checked.set_active (gtk.TRUE)
			checked.connect ("activate", self.sort_directories_first)

			menu = items.get_widget ("<popup>")
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def upload_selected (self, action, button):

		files = [ ]

		for i in self.list.selection:
			entry = [ ]
			entry.append (self.list.get_pixtext (i, 0)[0])
			# Get remaining entries
			for j in range (self.list.columns - 1):
				entry.append (self.list.get_text (i, j + 1))

			# Check if the file already exists remotely. If so replace
			# the file's size with the remote size. Later we'll check
			# if the size is smaller and perform a resume if it is
			remote_size = main.app.get_file_size (entry[0])
			if remote_size is not None:
				entry[4] = remote_size

			if entry[0] == "..":
				continue
			else:
				files.append (tuple (entry))

		self.list.unselect_all ()
		main.app.transfer (main.app.get_remote_dir (), self.get_dir (), files, ftp.UPLOAD, main.app['transfer_type'])

	def cd_selected (self, action, button):

		browse_win = browsewin.LocalCWD (self.get_dir ())

	def open_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			path = os.path.join (self.get_dir (), file)

			# Skip the home directory entry
			if file == ".." or \
			   not os.path.isfile (path) or \
			   not os.access (path, os.X_OK):
				continue
			else:
				# Execing code will go here
				pass

	def edit_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			path = os.path.join (self.get_dir (), file)

			# Editing code will go here

	def delete_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			type = self.list.get_text (i, 3)[0]
			path = os.path.join (self.dir, file)
			if type in '-lf':
				os.remove (path)
			else:
				os.rmdir (path)
		self.list.unselect_all ()
		self.update_listing (self.dir)

	def rename_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			orig_path = os.path.join (self.get_dir (), file)
			new = GtkExtra.input_box (title=_("ftpcube - Rename"), message=_("Rename %s to:") %file)
			if new:
				new_path = os.path.join (self.get_dir (), new)
				try:
					os.rename (orig_path, new_path)
				except OSError, strerror:
					GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube error: %s") %strerror, buttons=( [ _("Ok") ]))
					return
		self.update_listing (self.dir)

	def create_directory (self, action, button):

		dir = GtkExtra.input_box (title=_("ftpcube - Create Directory"), message=_("Directory Name:"))
		if dir is not None:
			path = os.path.join (self.get_dir (), dir)
			try:
				os.mkdir (path)
			except OSError, strerror:
				GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube error: %s") %strerror, buttons=( [ _("Ok") ]))
				return
			self.update_listing (self.dir)

	def update_status_dir (self, dir):
		self.status_label.set_text (dir)

	def enable_dir_update (self, button):
		pass

	def read_dir (self, dir):

		# If we say '.', we really mean our current directory
		if dir == '.':
			dir = self.dir

		try:
			files = os.listdir (dir)
		except OSError, (errno, strerror):
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
			return

		if self.hidden_files:
			files = filter (lambda x, self=self: not self.hidden_re.match (x), files)

		# Add the special file ".."
		files.insert (0, "..")

		# Now grab the file info and load it into the load
		listitems = []
		for f in files:

			# Attempt to retrieve the file and modification times
			try:
				# lstat is preferrable for systems that support it
				if os.name == 'posix':
					(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
					 st_atime, st_mtime, st_ctime) = os.lstat (os.path.join (dir, f))
				else:
					(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
					 st_atime, st_mtime, st_ctime) = os.stat (os.path.join (dir, f))
			except OSError, strerror:
				# Set the size and mtime to 0 then and carry on if we can't stat
				# the file for some reason but know it's there
				st_size = 0
				st_mtime = 0

			# Beautify the size string
			size = app.beautify_size (st_size)

			# Beautify the time
			match = self.time_re.match (time.ctime (st_mtime))
			mtime = match.group (1)

			# Set the appropriate file type
			if os.path.islink (os.path.join (self.dir, f)):
				type = "l"
			elif os.path.isdir (os.path.join (self.dir, f)):
				type = "d"
			else:
				type = "f"

			listitems.append ((f, size, mtime, type, "%s" %st_size))

		return listitems

	def update_listing (self, dir):

		# Strip off any trailing '/' characters
		if dir[-1] == '/':
			dir = dir[:-1]

		# Construct our new path
		if dir == '..':
			self.dir = os.path.dirname (self.dir)
		elif dir == '.':
			# Don't modify the path
			pass
		else:
			newpath = os.path.join (self.dir, dir)
			if not os.path.isdir (newpath):
				return
			else:
				self.dir = newpath

		# Get the new listing
		listitems = self.read_dir (self.dir)

		# Now change the list to reflect the new listing. Notice that we
		# pause the visuals during the update so it doesn't look funny
		self.list.freeze ()
		self.clear_items ()
		if self.sorted:
			listitems = self.sort_list (listitems)
		for i in range (len (listitems)):
			self.list.append (listitems[i])
			listitem = listitems[i]
			pixmap, mask = self.select_pixmap (listitem)
			self.list.set_pixmap (i, 0, pixmap, mask)
			self.list.set_pixtext (i, 0, listitem[0], 4, pixmap, mask)
		self.list.thaw ()

		# Update the status
		self.update_status_dir (self.dir)

	def sort_list (self, list):

		if self.directories_first:
			dirs = filter (lambda i: i[3] == 'd', list)
			links = filter (lambda i: i[3] == 'l', list)
			files = filter (lambda i: i[3] in 'f-', list)

			list = [ ]
			list.extend (self.perform_sort (dirs))
			list.extend (self.perform_sort (links))
			list.extend (self.perform_sort (files))
		else:
			list = self.perform_sort (list)
		return list

class RemoteWin (AbstractFileWin):

	headings = [ _('Filename'), _('Size'), _('Date'), _('Flags'), _('Truesize') ]

	list_re = re.compile ('([A-Za-z-]+)\s+(\d+)\s+([A-Za-z0-9]+)\s+([A-Za-z0-9]+)\s+(\d+)\s+([A-Za-z]+\s+[0-9]+\s+[0-9:]+)\s+(.*)', re.IGNORECASE)
	link_re = re.compile ('(.*) -> (.*)')
	broken_link_re = re.compile ('.*-> (.*)')
	time_re = re.compile ('[A-Za-z]+\s+([A-Za-z]+\s+\d+\s+\d+:\d+):(\d+)\s+\d+')

	def __init__ (self, container):

		AbstractFileWin.__init__ (self, container, self.headings)

		self.status_label = gtk.GtkLabel ()
		self.status_label.show ()
		self.status_bar.pack_start (self.status_label, expand=GTK.FALSE, padding=6)

		# Set the list settings for our file window
		for i in range(len (self.headings)):
			self.list.set_column_width (i, 100)
			self.list.set_column_justification (i, gtk.JUSTIFY_LEFT)
		# Explicitly make the first column for the filename bigger
		self.list.set_column_width (0, 200)
		self.list.set_column_visibility (4, 0)
		# Make file sizes right justified
		self.list.set_column_justification (1, GTK.JUSTIFY_RIGHT)
		# So we can't click the titles as buttons
		self.list.column_titles_passive ()

		self.cache = cache.CacheTree ()

	def select_items (self, list, r, c, event):

		if hasattr (event, 'type'):
			# Check for double clicks
			if event.type == 5 and not main.app.main_thread.busy ():
				selected = list.get_pixtext (r, 0)[0]

				# Construct our new dir path
				if selected == "..":
					main.app.main_thread.cwd ('..')

					# Update the status directory
					text = self.get_dir ()
					index = string.rfind (text, "/")
					if index == 0:
						self.update_status_dir ("/")
					else:
						self.update_status_dir (text[:index])

					self.get_listing ()
				else:
					# Check if we have a directory
					f = list.get_text (r, 3)[0]
					if f in 'ld':
						# Update the status directory before the action
						# so we can roll back in the case of an exception
						self.update_status_dir (os.path.join (self.get_dir (), selected))

						main.app.main_thread.set_exception_handler (self.set_listing_callback)
						main.app.main_thread.cwd (selected)
						main.app.main_thread.set_exception_handler (self.set_listing_callback)
						self.get_listing ()

	def set_listing_callback (self, errmsg):

		gtk.threads_enter ()
		main.app.ftp_error_tell_user (errmsg)
		self.update_status_dir (self.get_dir ())
		gtk.threads_leave ()

	def popup_menu (self, list, event):

		if event.button == 3:
			opts = [
				(_("/Download Selected"),         None, self.download_selected, 0, ''),
				#("/View File",                 None, None, 0, ''),
				(_("/Rename Selected"),           None, self.rename_selected, 0, ''),
				(_("/Delete Selected"),           None, self.delete_selected, 0, ''),
				#("/Recursive Delete",          None, None, 0, ''),
				("/Sep1",                         None,    None, 0, '<Separator>'),
				(_("/Create Directory"),          None, self.mkdir_selected, 0, ''),
				(_("/Remove Directories"),        None, self.rmdir_selected, 0, ''),
				(_("/Change Directory"),          None, self.cd_selected, 0, ''),
				(_("/Refresh Cache"),             None, self.refresh_selected, 0, ''),
				#("/Sep2",                      None, None, 0, '<Separator>'),
				#("/Compare Directories",       None, None, 0, ''),
				#("/Add Bookmarks",             None, None, 0, ''),
				#("/Sep3",                      None, None, 0, '<Separator>'),
				#("/Set Attributes",            None, None, 0, ''),
				#("/Custom Commands",           None, None, 0, ''),
				#("/Directory Info",            None, None, 0, ''),
				(_("/Sep4"),                      None, None, 0, '<Separator>'),
				(_("/Sort Order"),                None, None, 0, '<Branch>'),
				(_("/Sort Order/Unsorted"),       None, self.unsorted, 0, ''),
				(_("/Sort Order/Sort by Name"),   None, self.sort_by_name, 0, ''),
				(_("/Sort Order/Sort by Size"),   None, self.sort_by_size, 0, ''),
				(_("/Sort Order/Sort by Date"),   None, self.sort_by_date, 0, ''),
				(_("/Sort Order/Sep5"),           None, None, 0, '<Separator>'),
				(_("/Sort Order/Order Directories-Links-Files"), None, None, 0, '<ToggleItem>')
			]

			ag = gtk.GtkAccelGroup()
			items = gtk.GtkItemFactory(gtk.GtkMenu.get_type(), "<popup>", ag)
			items.create_items (opts)

			checked = items.get_widget (_('/Sort Order/Order Directories-Links-Files'))
			if self.directories_first:
				checked.set_active (gtk.TRUE)
			checked.connect ("activate", self.sort_directories_first)

			menu = items.get_widget ("<popup>")
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def clear_cache (self):

		del self.cache
		self.cache = cache.CacheTree ()

	def get_listing (self):

		path_node = self.cache.find_node (self.get_dir ())
		if path_node is None:
			main.app.main_thread.list ()
		else:
			listing = path_node.data
			self.set_list (listing)

	def update_listing (self, listing):

		# Create the ".." entry
		match = self.time_re.match (time.ctime (time.time ()))
		localtime = match.group (1)
		dotdotentry = ('..', _('4.00 KBytes'), localtime, '', '0')
		listitems = [ dotdotentry ]

		cache_node = self.cache.add_node (self.get_dir ())

		# Update the last listing array and parse the entries to update
		# the remote listing
		for entry in listing:
			tuple = self.parse_list_entry (entry)
			if tuple is None:
				continue
			else:
				file, truesize, date, flags = tuple

			# Beautify the size string
			size = app.beautify_size (truesize)

			# Check if the file is a link and if so, clean up the file name
			if flags and flags[0] == 'l':
				match = self.link_re.match (file)
				if match is not None:
					file, dest = match.groups ()
				else:
					# Try the broken link match
					match = self.broken_link_re.match (file)
					if match is not None:
						file = match.group (1)
						file = os.path.basename (file)
					else:
						print "WARNING: Match failed for: %s" %file

			listitems.append ((file, size, date, flags, "%s" %truesize))

		# Now change the list to reflect the new listing. Notice that we
		# pause the visuals during the update so it doesn't look funny
		if self.sorted:
			listitems = self.sort_list (listitems)
		cache_node.data = listitems
		self.set_list (listitems)

	def get_status (self):
		return self.status_label.get ()

	def update_status_dir (self, dir):

		self.dir = dir
		self.update_status (self.dir)
		main.app.update_dir_entry_status (self.dir)

	def update_status (self, text):
		self.status_label.set_text (text)

	def parse_list_entry (self, entry):

		match = self.list_re.match (entry)
		if match is None:
			#print "WARNING: listing match failed for: %s" %entry
			return None
		else:
			flags, hardlinks, user, group, size, date, file =  match.groups ()
			size = int (size)
			return file, size, date, flags

	def download_selected (self, action, button):

		files = [ ]

		# Extract the data from the CList. Note we don't use
		# get_row_data because it appears to be broken
		for i in self.list.selection:
			entry = [ ]
			entry.append (self.list.get_pixtext (i, 0)[0])
			# Get rest of entries
			for j in range (self.list.columns - 1):
				entry.append (self.list.get_text (i, j + 1))

			# Discard the ".." entry
			if entry[0] == "..":
				continue

			# Fix up the type
			if not entry[3][0] == '-':
				entry[3] = entry [3][0]
			else:
				entry[3] = 'f'

			files.append (tuple (entry))

		self.list.unselect_all ()
		main.app.transfer (self.get_dir (), main.app.get_local_dir (), files, ftp.DOWNLOAD, main.app['transfer_type'])

	def rename_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			new_name = GtkExtra.input_box (title=_("ftpcube - Remote Rename"), message=_("Rename %s to:" %file))
			if new_name is None:
				break
			main.app.main_thread.rename (file, new_name)
		main.app.main_thread.list ()

	def delete_selected (self, action, button):

		for i in self.list.selection:
			file = self.list.get_pixtext (i, 0)[0]
			type = self.list.get_text (i, 3)[0]
			if type in 'l-':
				main.app.main_thread.delete (file)
			else:
				main.app.main_thread.rmdir (file)

		self.list.unselect_all ()
		main.app.main_thread.list ()

	def mkdir_selected (self, action, button):

		dir = GtkExtra.input_box (title=_("ftpcube - Create Remote Directory"), message=_("Directory Name:"))
		if dir is not None:
			main.app.main_thread.mkdir (dir)
			main.app.main_thread.list ()

	def rmdir_selected (self, action, button):

		for i in self.list.selection:
			dir = self.list.get_pixtext (i, 0)[0]
			type = self.list.get_text (i, 3)[0]
			if type == 'd':
				main.app.main_thread.rmdir (dir)
		main.app.main_thread.list ()

	def cd_selected (self, action, button):

		browse_win = browsewin.RemoteCWD (self.cache, self.get_dir ())

	def refresh_selected (self, action, button):

		self.cache.remove_node (self.get_dir ())
		self.get_listing ()

	def set_list (self, list):

		self.list.freeze ()
		self.clear_items ()
		for i in range (len (list)):
			self.list.append (list[i])
			listitem = list[i]
			pixmap, mask = self.select_pixmap (listitem)
			self.list.set_pixmap (i, 0, pixmap, mask)
			self.list.set_pixtext (i, 0, listitem[0], 4, pixmap, mask)
		self.list.thaw ()

	def get_list (self):

		list = [ ]
		for i in range (self.list.rows):
			entry = [ ]
			for j in range (self.list.columns):
				if j == 0:
					entry.append (self.list.get_pixtext (i, j)[0])
				else:
					entry.append (self.list.get_text (i, j))
			list.append (tuple (entry))
		return list

	def sort_list (self, list):

		if self.directories_first:
			dirs = filter (lambda i: not i[3] or i[3][0] == 'd', list)
			links = filter (lambda i: i[3] and i[3][0] == 'l', list)
			files = filter (lambda i: i[3] and i[3][0] in 'f-', list)

			list = [ ]
			list.extend (self.perform_sort (dirs))
			list.extend (self.perform_sort (links))
			list.extend (self.perform_sort (files))
		else:
			list = self.perform_sort (list)
		return list



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import GtkExtra
import os
import main

from intl import _

class AboutWindow:

	"""About Window class

	Implements the about window object, which provides a popup window
	with a brief description of the program. This object is accessable
	through the main menu."""

	def __init__ (self):
		"""Initializes the AboutWindow object and creates the about
		window popup."""

		GtkExtra.message_box (title=_("ftpcube - About"),
		                      message=self.about_text (),
		                      pixmap=os.path.join (main.ICONS_PREFIX, "ftpcube.xpm"),
		                      buttons=( [ _("Ok") ] ))

	def about_text (self):
		"""Returns the about message"""

		return \
_("""FtpCube - A graphical FTP Client\n
FtpCube is a graphical ftp client written in Python with GTK bindings, and based on LeechFTP by Jan Debis.
Feel free to visit the homepage at: http://ftpcube.sourceforge.net.\n
Michael Gilfix <mgilfix@eecs.tufts.edu>
Copyright (C) 2001
Licensing: ARTISTIC""")



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import GtkExtra
import time
import main

from intl import _

class Log:
	"""A logging object.

	This object provides a wrapper for textual logging of FtpCube's commands.
	Marks are inserted into the log and start up, as well as some beautifying
	of the log messages."""

	def __init__ (self, log_file, mark=1):
		"""Initializes the logging object by opening the log file and inserting
		a mark (timestamp) unless 'mark' is set to 0."""

		self.file_delim = '\n'

		try:
			self.log = open (log_file, "a")
		except IOError:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s"), button=( [ _("Ok") ]))
			self.log = None

		if mark:
			self.mark ()

	def __del__ (self):
		"""Gracefully closes the logfile upon destruction of the object."""

		if self.log is not None:
			self.log.close ()

	def set_file_delim (self, delim):
		"""Sets the end-of-line delimeter used. Normally, you'd want a '\n' for
		unix systems and a '\n\r' for DOS-based systems."""

		self.file_delim = delim

	def write (self, msg):
		"""Adds a timestamp to the 'msg' string and writes out the message
		to the log file."""

		if self.log is not None:
			log_time = "%s: " %(time.asctime (time.localtime (time.time ())))
			self.log.write (log_time + msg + self.file_delim)
			self.log.flush ()

	def write_blank (self):
		self.write ('')

	def mark (self):
		"""Inserts a mark or timestamp into the log file."""

		self.write_blank ()
		self.write (_("----- MARK -----"))
		self.write_blank ()



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk
import sys, os
import main

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class CmdWin:

	"""Command window class

	This object displays the command window within the alloted
	container, which is a scrolled, non-editable text window. The
	command-window also takes care of all the formatting and color
	coding of the displayed text."""

	def __init__ (self, container):
		"""Initializes and displays the command window. Text colors
		are initialized to 'None' by default."""

		self.format = {
			'local'  : _("LOCAL >  %s\n"),
			'remote' : _("REMOTE> %s\n"),
			'error'  : _("! - %s\n")
		}

		self.colors = {
			'local' : None,
			'remote' : None,
			'error' : None
		}

		self.gdk_colors = {
			'local' : None,
			'remote' : None,
			'error' : None
		}

		win = gtk.GtkScrolledWindow ()
		win.set_policy (GTK.POLICY_AUTOMATIC, GTK.POLICY_AUTOMATIC)
		win.show ()
		container.add (win)

		self.text = gtk.GtkText ()
		self.text.set_usize (-1, 150)
		self.text.set_editable (gtk.FALSE)
		self.text.show ()
		win.add (self.text)

		self.colormap = self.text.get_colormap()

	def display_banner (self):
		"""Prints the application greeting in the command window."""

		self.insert_color_text ('local', _('Welcome to FtpCube v0.3.2'))
		self.insert_color_text ('local', _('Copyright (C) 2001  Michael Gilfix'))
		self.insert_color_text ('local', _('=========================================='))
		self.insert_color_text ('local', '')

	def insert_color_text(self, kind, text):
		"""Reformats the 'text' string according to its type string 'kind' and
		prints it in the appropriate color."""

		gdk_color = self.get_gdk_color(kind)
		text = self.format[kind] % text
		self.text.insert (None, gdk_color, None, text)

	def get_gdk_color(self, kind):
		"""Returns the GDK color object associated with the color type string
		'kind'."""

		current_color = main.app[kind + '_color']

		if self.colors[kind] == current_color:
			gdk_color = self.gdk_colors[kind]
		else:
			rgbo = map (lambda c: int (c * 65535.0), current_color)
			gdk_color = apply (self.colormap.alloc, rgbo)
			self.colors[kind] = current_color
			self.gdk_colors[kind] = gdk_color

		return gdk_color
